# standard libraries
import gettext

# third party libraries
# None

# local libraries
from nion.ui import Application
from nion.ui import Declarative

# ui imports
from . import Bindings
from . import Buttons
from . import CheckBoxes
from . import ComboBoxes
from . import Compositions
from . import ComponentContent
from . import ComponentLayout
from . import ComponentPolymorphic
from . import ComponentStack
from . import Converters
from . import Groups
from . import LineEdits
from . import ProgressBars
from . import RadioButtons
from . import Sliders
from . import Stacks
from . import StatusBar
from . import Tabs

_ = gettext.gettext


class Handler:

    def __init__(self):
        self.page_stack = None
        self.page_list = None

    def select_page(self, widget, current_index):
        self.page_stack.current_index = current_index

    def create_handler(self, component_id: str=None, **kwargs):
        if self.page_list and component_id:
            for page_cls, page_id, page_title in self.page_list:
                if page_id == component_id:
                    return page_cls.Handler()
        return None


def main(args, bootstrap_args):
    ui = Declarative.DeclarativeUI()

    page_list = [
        (Buttons, "buttons", _("Buttons")),
        (CheckBoxes, "check_boxes", _("Check Boxes")),
        (ComboBoxes, "combo_boxes", _("Combo Boxes")),
        (Bindings, "bindings", _("Bindings")),
        (Compositions, "compositions", _("Compositions")),
        (Converters, "converters", _("Converters")),
        (Groups, "groups", _("Groups")),
        (LineEdits, "line_edits", _("Line Edits")),
        (ProgressBars, "progress_bars", _("Progress Bars")),
        (RadioButtons, "radio_buttons", _("Radio Buttons")),
        (Sliders, "sliders", _("Sliders")),
        (Stacks, "stacks", _("Stacks")),
        (StatusBar, "status_bar", _("Status Bar")),
        (Tabs, "tabs", _("Tabs")),
        (ComponentLayout, "component_layout", _("Component Layout")),
        (ComponentStack, "component_stack", _("Component Stack")),
        (ComponentContent, "component_content", _("Component Content")),
        (ComponentPolymorphic, "component_polymorphic", _("Component Polymorphic")),
        ]

    handler = Handler()
    handler.page_list = page_list

    resources = dict()

    pages = list()
    items = list()

    for page_cls, page_id, page_title in page_list:
        resources[page_id] = ui.define_component(content=page_cls.construct_ui(ui))
        instance = ui.create_component_instance(page_id)
        pages.append(ui.create_column(instance, ui.create_stretch()))
        items.append(page_title)

    chooser_combo_box = ui.create_combo_box(items=items, on_current_index_changed="select_page")

    page_stack = ui.create_stack(*pages, name="page_stack")

    page_group = ui.create_group(page_stack, margin=8)

    main_column = ui.create_column(chooser_combo_box, page_group, spacing=8)

    window = ui.create_window(main_column, title=_("UI Demo"), margin=12, resources=resources)

    return Application.run_window(args, bootstrap_args, window, handler)



"""
A basic class to serve as the document controller of a typical one window application.
"""
from __future__ import annotations

# standard libraries
import asyncio
import enum
import functools
import gettext
import logging
import pathlib
import typing
import weakref

# local libraries
from nion.utils import Event
from nion.utils import Geometry
from nion.utils import Process
from nion.ui import UserInterface

if typing.TYPE_CHECKING:
    import numpy
    from nion.ui import Application


_ = gettext.gettext


class ActionContext:
    def __init__(self, application: Application.BaseApplication, window: typing.Optional[Window], focus_widget: typing.Optional[UserInterface.Widget]):
        self.application = application
        self.window = window
        self.focus_widget = focus_widget
        self.event_type_str: typing.Optional[str] = None
        self.parameters: typing.Dict[str, typing.Any] = dict()


class ActionStatus(enum.IntEnum):
    FINISHED = 0
    MODAL = 1
    CANCELLED = 2
    PASS = 3


class ActionResult:
    def __init__(self, status: ActionStatus):
        self.status = status
        self.results: typing.Dict[str, typing.Any] = dict()


class ReportType(enum.IntEnum):
    NOTSET = 0
    DEBUG = 10
    INFO = 20
    WARNING = 30
    ERROR = 40
    CRITICAL = 50


class Report:
    def __init__(self, type: ReportType, message: str):
        self.type = type
        self.message = message

    @classmethod
    def from_log_record(cls, record: logging.LogRecord) -> Report:
        return Report(ReportType(record.levelno), record.getMessage())


class ActionProperty:
    pass


class ActionStringProperty(ActionProperty):
    def __init__(self, name: str):
        self.name = name


class ActionIntegerProperty(ActionProperty):
    def __init__(self, name: str):
        self.name = name


class Action:
    action_id: str = str()
    action_name: str = str()
    action_role: typing.Optional[str] = None
    action_summary: typing.Optional[str] = None
    action_description: typing.Optional[str] = None
    action_parameters: typing.List[ActionProperty] = list()

    def __init__(self):
        self.__reports: typing.List[Report] = list()

    @property
    def reports(self) -> typing.List[Report]:
        return self.__reports

    def clear(self) -> None:
        self.__reports = list()

    def event(self, context: ActionContext) -> ActionResult:
        """Handle the event."""
        ...

    def execute(self, context: ActionContext) -> ActionResult:
        """Execute the action with the context. No user interaction allowed. May be called from scripts."""
        raise NotImplementedError()

    def invoke(self, context: ActionContext) -> ActionResult:
        """Called to execute the action with the context. User interaction allowed. Typically calls `execute`."""
        return self.execute(context)

    def is_checked(self, context: ActionContext) -> bool:
        return False

    def is_enabled(self, context: ActionContext) -> bool:
        return True

    def get_action_name(self, context: ActionContext) -> str:
        return self.action_name

    def report(self, type: ReportType, message: str) -> None:
        self.__reports.append(Report(type, message))

    def log_record(self, record: logging.LogRecord) -> None:
        self.__reports.append(Report.from_log_record(record))

    def get_string_property(self, context: ActionContext, name: str) -> str:
        return str(context.parameters.get(name, str()))

    def set_string_property(self, context: ActionContext, name: str, value: str) -> None:
        context.parameters[name] = value

    def get_int_property(self, context: ActionContext, name: str) -> int:
        return int(context.parameters.get(name, 0))

    def set_int_property(self, context: ActionContext, name: str, value: int) -> None:
        context.parameters[name] = value


actions: typing.Mapping[str, Action] = dict()


def register_action(action: Action) -> None:
    assert action.action_id not in actions
    typing.cast(typing.MutableMapping[str, Action], actions)[action.action_id] = action


action_shortcuts: typing.Mapping[str, typing.Mapping] = dict()


def add_action_shortcut(action_id: str, action_context: str, key_sequence: str) -> None:
    typing.cast(typing.MutableMapping[str, typing.MutableMapping], action_shortcuts).setdefault(action_id, dict())[action_context] = key_sequence


def register_action_shortcuts(action_shortcuts: typing.Mapping) -> None:
    for action_id, action_shortcut_d in action_shortcuts.items():
        for action_context, key_sequence in action_shortcut_d.items():
            add_action_shortcut(action_id, action_context, key_sequence)


def get_action_id_for_key(context: str, key) -> typing.Optional[str]:
    for action_id, action_shortcut_d in action_shortcuts.items():
        for action_context, key_sequence in action_shortcut_d.items():
            if action_context == context and key.text == key_sequence:  # TODO: match the actual key sequence
                return action_id
    return None


class Window:
    count = 0  # useful for detecting leaks in tests

    def __init__(self, ui: UserInterface.UserInterface, app: typing.Optional[Application.BaseApplication] = None,
                 parent_window: Window = None, window_style: typing.Optional[str] = None,
                 persistent_id: typing.Optional[str] = None):
        Window.count += 1
        self.ui = ui
        self.parent_window = parent_window
        self.app: typing.Optional[Application.BaseApplication] = app or (parent_window.app if parent_window else None)
        self.on_close: typing.Optional[typing.Callable[[], None]] = None
        parent_ui_window = parent_window._document_window if parent_window else None
        self.__document_window = self.ui.create_document_window(parent_window=parent_ui_window)
        if window_style:
            self.__document_window.window_style = window_style
        self.__persistent_id = persistent_id
        self.__shown = False
        self.__request_close = False

        self.__dialogs: typing.List[weakref.ReferenceType] = list()

        self._window_close_event = Event.Event()

        self.__document_window.on_periodic = self.periodic
        self.__document_window.on_queue_task = self.queue_task
        self.__document_window.on_clear_queued_tasks = self.clear_queued_tasks
        self.__document_window.on_add_task = self.add_task
        self.__document_window.on_clear_task = self.clear_task
        self.__document_window.on_about_to_show = self.about_to_show
        self.__document_window.on_about_to_close = self.about_to_close
        self.__document_window.on_activation_changed = self.activation_changed
        self.__document_window.on_key_pressed = self.key_pressed
        self.__document_window.on_key_released = self.key_released
        self.__document_window.on_size_changed = self.size_changed
        self.__document_window.on_position_changed = self.position_changed
        self.__document_window.on_refocus_widget = self.refocus_widget
        self.__document_window.on_ui_activity = self._register_ui_activity
        self.__periodic_queue = Process.TaskQueue()
        self.__periodic_set = Process.TaskSet()
        self.__modal_actions: typing.List[Action] = list()

        # define old-style menu actions for backwards compatibility
        self._close_action = None
        self._page_setup_action = None
        self._print_action = None
        self._quit_action = None
        self._undo_action = None
        self._redo_action = None
        self._cut_action = None
        self._copy_action = None
        self._paste_action = None
        self._delete_action = None
        self._select_all_action = None
        self._minimize_action = None
        self._zoom_action = None
        self._bring_to_front_action = None

        # configure the event loop object
        logger = logging.getLogger()
        old_level = logger.level
        logger.setLevel(logging.INFO)
        self.__event_loop = asyncio.new_event_loop()  # outputs a debugger message!
        logger.setLevel(old_level)

        if app:
            app._window_created(self)

    def close(self) -> None:
        self.__request_close = False
        self._finish_periodic()  # required to finish periodic operations during tests
        self._close_dialogs()
        self._window_close_event.fire(self)
        self._window_close_event = typing.cast(Event.Event, None)
        self.on_close = None
        Process.close_event_loop(self.__event_loop)
        self.__event_loop = typing.cast(asyncio.AbstractEventLoop, None)
        self.ui.destroy_document_window(self.__document_window)  # close the ui window
        self.__document_window = typing.cast(UserInterface.Window, None)
        self.__periodic_queue = typing.cast(Process.TaskQueue, None)
        self.__periodic_set = typing.cast(Process.TaskSet, None)
        self.__modal_actions = typing.cast(typing.List[Action], None)
        self._close_action = None
        self._page_setup_action = None
        self._print_action = None
        self._quit_action = None
        self._undo_action = None
        self._redo_action = None
        self._cut_action = None
        self._copy_action = None
        self._paste_action = None
        self._delete_action = None
        self._select_all_action = None
        self._minimize_action = None
        self._zoom_action = None
        self._bring_to_front_action = None
        self.parent_window = None
        self.app = None
        Window.count -= 1

    @property
    def _document_window(self) -> UserInterface.Window:
        # for testing only
        return self.__document_window

    def _create_menus(self) -> None:
        menu_descriptions = [
            {"type": "menu", "menu_id": "file", "title": _("File"), "items":
                [
                    {"type": "item", "action_id": "window.close"},
                    {"type": "separator"},
                    {"type": "item", "action_id": "window.page_setup"},
                    {"type": "item", "action_id": "window.print"},
                    {"type": "separator"},
                    {"type": "item", "action_id": "application.exit"},
                ]
             },
            {"type": "menu", "menu_id": "edit", "title": _("Edit"), "items":
                [
                    {"type": "item", "action_id": "window.undo"},
                    {"type": "item", "action_id": "window.redo"},
                    {"type": "separator"},
                    {"type": "item", "action_id": "window.cut"},
                    {"type": "item", "action_id": "window.copy"},
                    {"type": "item", "action_id": "window.paste"},
                    {"type": "item", "action_id": "window.delete"},
                    {"type": "item", "action_id": "window.select_all"},
                ]
             },
            {"type": "menu", "menu_id": "window", "title": _("Window"), "items":
                [
                    {"type": "item", "action_id": "window.minimize"},
                    {"type": "separator"},
                    {"type": "item", "action_id": "window.zoom"},
                    {"type": "item", "action_id": "window.bring_to_front"},
                ]
             },
            {"type": "menu", "menu_id": "help", "title": _("Help"), "items":
                [
                ]
             },
        ]

        self.build_menu(None, typing.cast(typing.List[typing.Dict], menu_descriptions))

    def _adjust_menus(self) -> None:
        # called when key may be shortcut. does not work for sub-menus.
        for menu in self.__document_window.menus:
            self._menu_about_to_show(menu)

    def _request_exit(self) -> None:
        if self.app:
            self.app.exit()

    @property
    def window_file_path(self) -> typing.Optional[pathlib.Path]:
        return self.__document_window.window_file_path

    @window_file_path.setter
    def window_file_path(self, value: typing.Optional[pathlib.Path]) -> None:
        self.__document_window.window_file_path = value

    def request_close(self) -> None:
        self.__document_window.request_close()

    def queue_request_close(self) -> None:
        # used to request a close from inside an event loop task.
        # sets the flag and the request close will be executed during the next periodic.
        self.__request_close = True

    def _register_ui_activity(self) -> None:
        pass

    def _finish_periodic(self) -> None:
        # recognize when we're running as test and finish out periodic operations
        if not self.__document_window.has_event_loop:
            self.periodic()

    def periodic(self) -> None:
        self.__periodic_queue.perform_tasks()
        self.__periodic_set.perform_tasks()
        self.__event_loop.stop()
        self.__event_loop.run_forever()
        if self.app:
            self.app.periodic()
        # if the request close flag is set, request the close.
        # this must be at the end of periodic since it will result in the window
        # close method being called.
        if self.__request_close:
            self.request_close()

    def exec_action_events(self, event_type_str: str, **kwargs) -> bool:
        action_context = None
        for action in self.__modal_actions:
            if not action_context:
                action_context = self._get_action_context()
                action_context.event_type_str = event_type_str
                for k, v in kwargs.items():
                    setattr(action_context, k, v)
            action_result = action.event(action_context)
            # finished or cancelled will remove the action
            if action_result.status not in {ActionStatus.MODAL, ActionStatus.PASS}:
                self.__modal_actions.remove(action)
            # modal, finished, or cancelled will stop iterating and return True
            if action_result.status != ActionStatus.PASS:
                for report in action.reports:
                    self.display_report(report)
                return True
        # pass will return False
        return False

    def _close_dialogs(self) -> None:
        for weak_dialog in self.__dialogs:
            dialog = typing.cast("Window", weak_dialog())
            if dialog:
                try:
                    dialog.request_close()
                except Exception:
                    pass
        self.__dialogs = list()

    def is_dialog_type_open(self, dialog_class: typing.Type[Window]) -> bool:
        for dialog_weakref in self.__dialogs:
            if isinstance(dialog_weakref(), dialog_class):
                return True
        return False

    def register_dialog(self, dialog: Window) -> None:
        old_on_close = dialog.on_close

        def close_dialog():
            self.__dialogs.remove(weakref.ref(dialog))
            if callable(old_on_close):
                old_on_close()

        dialog.on_close = close_dialog
        self.__dialogs.append(weakref.ref(dialog))

    @property
    def event_loop(self) -> asyncio.AbstractEventLoop:
        return self.__event_loop

    def attach_widget(self, widget: UserInterface.Widget) -> None:
        self.__document_window.attach(widget)

    def detach_widget(self) -> None:
        self.__document_window.detach()

    def about_to_show(self) -> None:
        if self.__persistent_id:
            geometry = self.ui.get_persistent_string("{}/Geometry".format(self.__persistent_id))
            state = self.ui.get_persistent_string("{}/State".format(self.__persistent_id))
            self.restore(geometry, state)
        self.__shown = True

    def about_to_close(self, geometry: str, state: str) -> None:
        # this method is invoked when the low level window is about to close.
        # subclasses can override this method to save geometry and state.
        if callable(self.on_close):
            self.on_close()
        # this call will close this object and subsequently the ui window, but not the low level window itself.
        # it will, however, delete the widget hierarchy as a consequence of closing the ui window.
        # so care must be taken to not close windows in cases where the widget triggering the close is still in use.
        self.close()

    def refocus_widget(self, widget: UserInterface.Widget) -> None:
        widget.refocus()

    def __save_bounds(self) -> None:
        if self.__shown and self.__persistent_id:
            geometry, state = self.save()
            self.ui.set_persistent_string("{}/Geometry".format(self.__persistent_id), geometry)
            self.ui.set_persistent_string("{}/State".format(self.__persistent_id), state)

    def activation_changed(self, activated: bool) -> None:
        pass

    def size_changed(self, width: int, height: int) -> None:
        self.__save_bounds()

    def position_changed(self, x: int, y: int) -> None:
        self.__save_bounds()

    def key_pressed(self, key: UserInterface.Key) -> bool:
        return False

    def key_released(self, key: UserInterface.Key) -> bool:
        if key.modifiers.control and key.key:
            self._adjust_menus()
        return False

    def drag(self, mime_data: UserInterface.MimeData, thumbnail: numpy.ndarray, hot_spot_x: int, hot_spot_y: int) -> None:
        self.__document_window.root_widget.drag(mime_data, thumbnail, hot_spot_x, hot_spot_y)

    @property
    def title(self) -> str:
        return self.__document_window.title

    @title.setter
    def title(self, value: str) -> None:
        self.__document_window.title = value

    def get_file_paths_dialog(self, title: str, directory: str, filter: str, selected_filter: str = None) -> typing.Tuple[typing.List[str], str, str]:
        return self.__document_window.get_file_paths_dialog(title, directory, filter, selected_filter)

    def get_file_path_dialog(self, title: str, directory: str, filter: str, selected_filter: str = None) -> typing.Tuple[typing.List[str], str, str]:
        return self.__document_window.get_file_path_dialog(title, directory, filter, selected_filter)

    def get_save_file_path(self, title: str, directory: str, filter: str, selected_filter: str = None) -> typing.Tuple[str, str, str]:
        return self.__document_window.get_save_file_path(title, directory, filter, selected_filter)

    def create_dock_widget(self, widget: UserInterface.Widget, panel_id: str, title: str,
                           positions: typing.Sequence[str], position: str) -> UserInterface.DockWidget:
        return self.__document_window.create_dock_widget(widget, panel_id, title, positions, position)

    def tabify_dock_widgets(self, dock_widget1: UserInterface.DockWidget, dock_widget2: UserInterface.DockWidget) -> None:
        self.__document_window.tabify_dock_widgets(dock_widget1, dock_widget2)

    @property
    def screen_size(self) -> Geometry.IntSize:
        return self.__document_window.screen_size

    @property
    def screen_logical_dpi(self) -> float:
        return self.__document_window.screen_logical_dpi

    @property
    def screen_physical_dpi(self) -> float:
        return self.__document_window.screen_physical_dpi

    @property
    def display_scaling(self) -> float:
        return self.__document_window.display_scaling

    def get_font_metrics(self, font: str, text: str) -> UserInterface.FontMetrics:
        return self.ui.get_font_metrics(font, text)

    @property
    def focus_widget(self) -> typing.Optional[UserInterface.Widget]:
        focus_widget = self.__document_window.focus_widget
        if focus_widget:
            return focus_widget
        for dock_widget in self.dock_widgets:
            focus_widget = dock_widget.focus_widget
            if focus_widget:
                return focus_widget
        return None

    @property
    def dock_widgets(self) -> typing.Sequence[UserInterface.DockWidget]:
        return self.__document_window.dock_widgets

    def show(self, *, size: Geometry.IntSize = None, position: Geometry.IntPoint = None) -> None:
        self.__document_window.show(size=size, position=position)

    def add_menu(self, title: str, menu_id: str = None) -> UserInterface.Menu:
        return self.__document_window.add_menu(title, menu_id)

    def insert_menu(self, title: str, before_menu: UserInterface.Menu, menu_id: str = None) -> UserInterface.Menu:
        return self.__document_window.insert_menu(title, before_menu, menu_id)

    def create_sub_menu(self, title: str = None, menu_id: str = None) -> UserInterface.Menu:
        return self.ui.create_sub_menu(self.__document_window, title, menu_id)

    def create_context_menu(self) -> UserInterface.Menu:
        return self.ui.create_context_menu(self.__document_window)

    def restore(self, geometry: str, state: str) -> None:
        self.__document_window.restore(geometry, state)

    def save(self) -> typing.Tuple[str, str]:
        return self.__document_window.save()

    # tasks can be added in two ways, queued or added
    # queued tasks are guaranteed to be executed in the order queued.
    # added tasks are only executed if not replaced before execution.
    # added tasks do not guarantee execution order or execution at all.

    def add_task(self, key: str, task: typing.Callable[[], None]) -> None:
        assert task
        self.__periodic_set.add_task(key + str(id(self)), task)

    def clear_task(self, key: str) -> None:
        self.__periodic_set.clear_task(key + str(id(self)))

    def queue_task(self, task: typing.Callable[[], None]) -> None:
        assert task
        self.__periodic_queue.put(task)

    def clear_queued_tasks(self) -> None:
        self.__periodic_queue.clear_tasks()

    def handle_quit(self) -> None:
        assert self.app
        self.app.exit()

    def _dispatch_any_to_focus_widget(self, method: str, *args, **kwargs) -> bool:
        focus_widget = self.focus_widget
        if focus_widget and focus_widget._dispatch_any(method, *args, **kwargs):
            return True
        if hasattr(self, method) and getattr(self, method)(*args, **kwargs):
            return True
        return False

    def _can_dispatch_to_focus_widget(self, method: str) -> bool:
        focus_widget = self.focus_widget
        if focus_widget and focus_widget._can_dispatch_any(method):
            return True
        if hasattr(self, method):
            return True
        return False

    def build_menu(self, menu: typing.Optional[UserInterface.Menu], menu_descriptions: typing.Sequence[typing.Mapping]) -> None:
        for item_d in menu_descriptions:
            item_type = item_d["type"]
            if item_type == "menu":
                assert menu is None
                menu_id = item_d["menu_id"]
                menu_title = item_d["title"]
                menu_items = item_d["items"]
                new_menu = self.__document_window.get_menu(menu_id)
                if not new_menu:
                    new_menu = self.add_menu(menu_title, menu_id)
                    new_menu.on_about_to_show = functools.partial(self._menu_about_to_show, new_menu)
                setattr(self, "_" + menu_id + "_menu", new_menu)
                self.build_menu(new_menu, menu_items)
            elif item_type == "item":
                action_id = item_d["action_id"]
                action = actions.get(action_id)
                if action:
                    key_sequence = action_shortcuts.get(action_id, dict()).get("window")
                    role = getattr(action, "about_role", None)
                    assert menu is not None
                    menu.add_menu_item(action.action_name, functools.partial(self.perform_action, action_id),
                                       key_sequence=key_sequence, role=role, action_id=action_id)
                else:
                    logging.debug("Unregistered action {action_id}")
            elif item_type == "separator":
                assert menu is not None
                menu.add_separator()
            elif item_type == "sub_menu":
                assert menu is not None
                menu_id = item_d["menu_id"]
                menu_title = item_d["title"]
                menu_items = item_d["items"]
                new_menu = self.create_sub_menu(menu_title, menu_id)
                menu.add_sub_menu(menu_title, new_menu)
                new_menu.on_about_to_show = functools.partial(self._menu_about_to_show, new_menu)
                # setattr(self, "_" + menu_id + "_menu", new_menu)
                self.build_menu(new_menu, menu_items)

    def _get_action_context(self) -> ActionContext:
        assert self.app
        focus_widget = self.focus_widget
        return ActionContext(self.app, self, focus_widget)

    def _apply_menu_state(self, action_id: str, action_context: ActionContext) -> None:
        menu_action = self.__document_window.get_menu_action(action_id)
        if menu_action and menu_action.action_id:
            action = actions.get(menu_action.action_id)
            if action:
                title = action.get_action_name(action_context)
                enabled = action and action.is_enabled(action_context)
                checked = action and action.is_checked(action_context)
                menu_action.apply_state(UserInterface.MenuItemState(title=title, enabled=enabled, checked=checked))

    def is_action_enabled(self, action_id: str, action_context: ActionContext) -> bool:
        action = actions.get(action_id)
        if action and action.is_enabled(action_context):
            return True
        return False

    def add_action_to_menu_if_enabled(self, menu: UserInterface.Menu, action_id: str, action_context: ActionContext) -> typing.Optional[Action]:
        action = actions.get(action_id)
        if action and action.is_enabled(action_context):
            self.add_action_to_menu(menu, action_id, action_context)
        return action

    def add_action_to_menu(self, menu: UserInterface.Menu, action_id: str, action_context: ActionContext) -> typing.Optional[Action]:
        action = actions.get(action_id)
        if action:
            key_sequence = action_shortcuts.get(action_id, dict()).get("window")
            assert menu is not None

            def perform_action():
                self.perform_action_in_context(action_id, action_context)

            def queue_perform_action():
                # delay execution to ensure menu closes properly
                # this would manifest itself by export dialog crashes in nionswift.
                self.queue_task(perform_action)

            menu_action = menu.add_menu_item(action.action_name,
                                             queue_perform_action,
                                             key_sequence=key_sequence, action_id=action_id)
            title = action.get_action_name(action_context)
            enabled = action and action.is_enabled(action_context)
            checked = action and action.is_checked(action_context)
            menu_action.apply_state(UserInterface.MenuItemState(title=title, enabled=enabled, checked=checked))
        return action

    def execute_action(self, action_id: str, action_context: typing.Optional[ActionContext] = None,
                       parameters: typing.Optional[typing.List] = None) -> ActionResult:
        context = action_context or self._get_action_context()
        if parameters:
            context.parameters.update(parameters)
        action = actions[action_id]
        return action.execute(context)

    def perform_action(self, action_id: str) -> None:
        self.perform_action_in_context(action_id, self._get_action_context())

    def perform_action_in_context(self, action_id: str, action_context: ActionContext) -> None:
        action = actions.get(action_id)
        if action and action not in self.__modal_actions:
            action.clear()
            if action.invoke(action_context).status == ActionStatus.MODAL:
                self.__modal_actions.append(action)
            for report in action.reports:
                self.display_report(report)

    def display_report(self, report: Report) -> None:
        from nion.ui import Dialog  # avoid circular reference
        if report.type == ReportType.DEBUG:
            logging.debug(report.message)
        elif report.type == ReportType.INFO:
            logging.info(report.message)
        elif report.type == ReportType.WARNING:
            logging.warning(report.message)
            Dialog.NotificationDialog(self.ui, message=f"WARNING: {report.message}", parent_window=self).show()
        elif report.type == ReportType.ERROR:
            logging.error(report.message)
            Dialog.NotificationDialog(self.ui, message=f"ERROR: {report.message}", parent_window=self).show()

    def display_log_record(self, record: logging.LogRecord) -> None:
        self.display_report(Report.from_log_record(record))

    def _get_menu_item_state(self, command_id: str) -> typing.Optional[UserInterface.MenuItemState]:
        # if there is a specific menu item state for the command_id, use it
        # otherwise, if the handle method exists, return an enabled menu item
        # otherwise, don't handle
        handle_method = "handle_" + command_id
        menu_item_state_method = "get_" + command_id + "_menu_item_state"
        if hasattr(self, menu_item_state_method):
            menu_item_state = getattr(self, menu_item_state_method)()
            if menu_item_state:
                return menu_item_state
        if hasattr(self, handle_method):
            return UserInterface.MenuItemState(title=None, enabled=True, checked=False)
        return None

    def _get_focus_widget_menu_item_state(self, command_id: str) -> typing.Optional[UserInterface.MenuItemState]:
        focus_widget = self.focus_widget
        if focus_widget:
            menu_item_state = focus_widget._get_menu_item_state(command_id)
            if menu_item_state:
                return menu_item_state
        return self._get_menu_item_state(command_id)

    # standard menu items

    def _menu_about_to_show(self, menu: UserInterface.Menu) -> None:
        if self.app and self.app._menu_about_to_show(self, menu):
            pass
        elif menu.menu_id == "file":
            self._file_menu_about_to_show()
        elif menu.menu_id == "edit":
            self._edit_menu_about_to_show()
        elif menu.menu_id == "window":
            self._window_menu_about_to_show()
        # perform enable/disable/title for all menus
        action_context = self._get_action_context()
        for menu_action in menu.get_menu_actions():
            if menu_action.action_id:
                action = actions.get(menu_action.action_id)
                if action:
                    title = action.get_action_name(action_context)
                    enabled = action and action.is_enabled(action_context)
                    checked = action and action.is_checked(action_context)
                    menu_action.apply_state(UserInterface.MenuItemState(title=title, enabled=enabled, checked=checked))

    def _file_menu_about_to_show(self) -> None:
        action_context = self._get_action_context()
        self._apply_menu_state("window.close", action_context)
        self._apply_menu_state("window.page_setup", action_context)
        self._apply_menu_state("window.print", action_context)
        self._apply_menu_state("application.exit", action_context)
        # handle old style for backwards compatibility
        if self._close_action:
            self._close_action.enabled = True
        if self._page_setup_action:
            self._page_setup_action.apply_state(self._get_focus_widget_menu_item_state("page_setup"))
        if self._print_action:
            self._print_action.apply_state(self._get_focus_widget_menu_item_state("print"))
        if self._quit_action:
            self._quit_action.enabled = True

    def _edit_menu_about_to_show(self) -> None:
        action_context = self._get_action_context()
        self._apply_menu_state("window.undo", action_context)
        self._apply_menu_state("window.redo", action_context)
        self._apply_menu_state("window.cut", action_context)
        self._apply_menu_state("window.copy", action_context)
        self._apply_menu_state("window.paste", action_context)
        self._apply_menu_state("window.delete", action_context)
        self._apply_menu_state("window.select_all", action_context)
        # handle old style for backwards compatibility
        if self._undo_action:
            self._undo_action.apply_state(self._get_focus_widget_menu_item_state("undo"))
        if self._redo_action:
            self._redo_action.apply_state(self._get_focus_widget_menu_item_state("redo"))
        if self._cut_action:
            self._cut_action.apply_state(self._get_focus_widget_menu_item_state("cut"))
        if self._copy_action:
            self._copy_action.apply_state(self._get_focus_widget_menu_item_state("copy"))
        if self._paste_action:
            self._paste_action.apply_state(self._get_focus_widget_menu_item_state("paste"))
        if self._delete_action:
            self._delete_action.apply_state(self._get_focus_widget_menu_item_state("delete"))
        if self._select_all_action:
            self._select_all_action.apply_state(self._get_focus_widget_menu_item_state("select_all"))

    def _window_menu_about_to_show(self) -> None:
        action_context = self._get_action_context()
        self._apply_menu_state("window.minimize", action_context)
        self._apply_menu_state("window.zoom", action_context)
        self._apply_menu_state("window.bring_to_front", action_context)
        # handle old style for backwards compatibility
        if self._minimize_action:
            self._minimize_action.apply_state(self._get_focus_widget_menu_item_state("minimize"))
        if self._zoom_action:
            self._zoom_action.apply_state(self._get_focus_widget_menu_item_state("zoom"))
        if self._bring_to_front_action:
            self._bring_to_front_action.apply_state(self._get_focus_widget_menu_item_state("bring_to_front"))

    def _page_setup(self) -> None:
        self.perform_action("window.page_setup")

    def _print(self) -> None:
        self.perform_action("window.print")

    def _cut(self) -> None:
        self._dispatch_any_to_focus_widget("handle_cut")

    def _copy(self) -> None:
        self._dispatch_any_to_focus_widget("handle_copy")

    def _paste(self) -> None:
        self._dispatch_any_to_focus_widget("handle_paste")

    def _delete(self) -> None:
        self._dispatch_any_to_focus_widget("handle_delete")

    def _select_all(self) -> None:
        self._dispatch_any_to_focus_widget("handle_select_all")

    def _undo(self) -> None:
        self._dispatch_any_to_focus_widget("handle_undo")

    def _redo(self) -> None:
        self._dispatch_any_to_focus_widget("handle_redo")

    def _minimize(self) -> None:
        self._dispatch_any_to_focus_widget("handle_minimize")

    def _zoom(self) -> None:
        self._dispatch_any_to_focus_widget("handle_zoom")

    def _bring_to_front(self) -> None:
        self._dispatch_any_to_focus_widget("bring_to_front")


class AboutBoxAction(Action):
    action_id = "application.about"
    action_name = _("About...")
    action_role = "about"

    def execute(self, context: ActionContext) -> ActionResult:
        raise NotImplementedError()

    def invoke(self, context: ActionContext) -> ActionResult:
        show_about_box = getattr(context.window, "show_about_box", None)
        show_about_box = show_about_box or getattr(context.application, "show_about_box", None)
        if callable(show_about_box):
            show_about_box()
        return ActionResult(ActionStatus.FINISHED)


class BringToFrontAction(Action):
    action_id = "window.bring_to_front"
    action_name = _("Bring to Front")

    def execute(self, context: ActionContext) -> ActionResult:
        if context.window:
            context.window._dispatch_any_to_focus_widget("bring_to_front")
        return ActionResult(ActionStatus.FINISHED)

    def is_enabled(self, context: ActionContext) -> bool:
        return hasattr(context.window, "bring_to_front")


class CloseWindowAction(Action):
    action_id = "window.close"
    action_name = _("Close Window")

    def execute(self, context: ActionContext) -> ActionResult:
        if context.window:
            context.window.request_close()
        return ActionResult(ActionStatus.FINISHED)


class CopyAction(Action):
    action_id = "window.copy"
    action_name = _("Copy")

    def execute(self, context: ActionContext) -> ActionResult:
        if context.window:
            context.window._dispatch_any_to_focus_widget("handle_copy")
        return ActionResult(ActionStatus.FINISHED)

    def is_enabled(self, context: ActionContext) -> bool:
        return hasattr(context.window, "handle_copy")


class CutAction(Action):
    action_id = "window.cut"
    action_name = _("Cut")

    def execute(self, context: ActionContext) -> ActionResult:
        if context.window:
            context.window._dispatch_any_to_focus_widget("handle_cut")
        return ActionResult(ActionStatus.FINISHED)

    def is_enabled(self, context: ActionContext) -> bool:
        return hasattr(context.window, "handle_cut")


class DeleteAction(Action):
    action_id = "window.delete"
    action_name = _("Delete")

    def execute(self, context: ActionContext) -> ActionResult:
        if context.window:
            context.window._dispatch_any_to_focus_widget("handle_delete")
        return ActionResult(ActionStatus.FINISHED)

    def is_enabled(self, context: ActionContext) -> bool:
        return hasattr(context.window, "handle_delete")


class ExitAction(Action):
    action_id = "application.exit"
    action_name = _("Exit")
    action_role = "quit"

    def execute(self, context: ActionContext) -> ActionResult:
        context.application.exit()
        return ActionResult(ActionStatus.FINISHED)


class MinimizeAction(Action):
    action_id = "window.minimize"
    action_name = _("Minimize")

    def execute(self, context: ActionContext) -> ActionResult:
        if context.window:
            context.window._dispatch_any_to_focus_widget("handle_minimize")
        return ActionResult(ActionStatus.FINISHED)

    def is_enabled(self, context: ActionContext) -> bool:
        return hasattr(context.window, "handle_minimize")


class PageSetupAction(Action):
    action_id = "window.page_setup"
    action_name = _("Page Setup")

    def execute(self, context: ActionContext) -> ActionResult:
        raise NotImplementedError()

    def invoke(self, context: ActionContext) -> ActionResult:
        if context.window:
            context.window._dispatch_any_to_focus_widget("handle_page_setup")
        return ActionResult(ActionStatus.FINISHED)

    def is_enabled(self, context: ActionContext) -> bool:
        return hasattr(context.window, "handle_page_setup")


class PasteAction(Action):
    action_id = "window.paste"
    action_name = _("Paste")

    def execute(self, context: ActionContext) -> ActionResult:
        if context.window:
            context.window._dispatch_any_to_focus_widget("handle_paste")
        return ActionResult(ActionStatus.FINISHED)

    def is_enabled(self, context: ActionContext) -> bool:
        return hasattr(context.window, "handle_paste")


class PreferencesAction(Action):
    action_id = "application.preferences"
    action_name = _("Preferences...")
    action_role = "preferences"

    def execute(self, context: ActionContext) -> ActionResult:
        raise NotImplementedError()

    def invoke(self, context: ActionContext) -> ActionResult:
        open_preferences = getattr(context.window, "open_preferences", None)
        open_preferences = open_preferences or getattr(context.application, "open_preferences", None)
        if callable(open_preferences):
            open_preferences()
        return ActionResult(ActionStatus.FINISHED)


class PrintAction(Action):
    action_id = "window.print"
    action_name = _("Print...")

    def execute(self, context: ActionContext) -> ActionResult:
        raise NotImplementedError()

    def invoke(self, context: ActionContext) -> ActionResult:
        if context.window:
            context.window._dispatch_any_to_focus_widget("handle_print")
        return ActionResult(ActionStatus.FINISHED)

    def is_enabled(self, context: ActionContext) -> bool:
        return hasattr(context.window, "handle_print")


class RedoAction(Action):
    action_id = "window.redo"
    action_name = _("Redo")

    def execute(self, context: ActionContext) -> ActionResult:
        if context.window:
            context.window._dispatch_any_to_focus_widget("handle_redo")
        return ActionResult(ActionStatus.FINISHED)

    def is_enabled(self, context: ActionContext) -> bool:
        return hasattr(context.window, "handle_redo")


class SelectAllAction(Action):
    action_id = "window.select_all"
    action_name = _("Select All")

    def execute(self, context: ActionContext) -> ActionResult:
        if context.window:
            context.window._dispatch_any_to_focus_widget("handle_select_all")
        return ActionResult(ActionStatus.FINISHED)

    def is_enabled(self, context: ActionContext) -> bool:
        if context.window:
            return context.window._can_dispatch_to_focus_widget("handle_select_all") or hasattr(context.window, "handle_select_all")
        return False


class UndoAction(Action):
    action_id = "window.undo"
    action_name = _("Undo")

    def execute(self, context: ActionContext) -> ActionResult:
        if context.window:
            context.window._dispatch_any_to_focus_widget("handle_undo")
        return ActionResult(ActionStatus.FINISHED)

    def is_enabled(self, context: ActionContext) -> bool:
        return hasattr(context.window, "handle_undo")


class ZoomAction(Action):
    action_id = "window.zoom"
    action_name = _("Zoom")

    def execute(self, context: ActionContext) -> ActionResult:
        if context.window:
            context.window._dispatch_any_to_focus_widget("handle_zoom")
        return ActionResult(ActionStatus.FINISHED)

    def is_enabled(self, context: ActionContext) -> bool:
        return hasattr(context.window, "handle_zoom")


register_action(AboutBoxAction())
register_action(BringToFrontAction())
register_action(CloseWindowAction())
register_action(CopyAction())
register_action(CutAction())
register_action(DeleteAction())
register_action(ExitAction())
register_action(MinimizeAction())
register_action(PageSetupAction())
register_action(PasteAction())
register_action(PreferencesAction())
register_action(PrintAction())
register_action(SelectAllAction())
register_action(UndoAction())
register_action(RedoAction())
register_action(ZoomAction())

action_shortcuts_dict = {
    "application.exit": {"window": "quit"},
    "window.close": {"window": "close"},
    "window.print": {"window": "Ctrl+P"},
    "window.undo": {"window": "undo"},
    "window.redo": {"window": "redo"},
    "window.cut": {"window": "cut"},
    "window.copy": {"window": "copy"},
    "window.paste": {"window": "paste"},
    "window.delete": {"window": "delete"},
    "window.select_all": {"window": "select-all"},
}

register_action_shortcuts(action_shortcuts_dict)



"""
    DrawingContext module contains classes related to drawing context.

    DrawingContexts are able to be handled directly by the UI system or
    produce javascript or svg to do the drawing.
"""

# standard libraries
import base64
import collections
from contextlib import contextmanager
import copy
import io
import logging
import math
import re
import struct
import sys
import time
import threading
import typing
import xml.sax.saxutils

# third party libraries
import imageio
import numpy

# local libraries
# None

# pylint: disable=star-args


def get_rgba_view_from_rgba_data(rgba_data):
    return rgba_data.view(numpy.uint8).reshape(rgba_data.shape + (4,))


def get_rgba_data_from_rgba(rgba_image):
    return rgba_image.view(numpy.uint32).reshape(rgba_image.shape[:-1])


def get_byte_view(rgba_image):
    return rgba_image.view(numpy.uint8).reshape(rgba_image.shape + (-1, ))


def get_red_view(rgba_image, byteorder=None):
    if byteorder is None:
        byteorder = sys.byteorder
    bytes = get_byte_view(rgba_image)
    assert bytes.shape[-1] == 4
    if byteorder == 'little':
        return bytes[..., 2]  # strip A off BGRA
    else:
        return bytes[..., 1]  # strip A off ARGB


def get_green_view(rgba_image, byteorder=None):
    if byteorder is None:
        byteorder = sys.byteorder
    bytes = get_byte_view(rgba_image)
    assert bytes.shape[-1] == 4
    if byteorder == 'little':
        return bytes[..., 1]  # strip A off BGRA
    else:
        return bytes[..., 2]  # strip A off ARGB


def get_blue_view(rgba_image, byteorder=None):
    if byteorder is None:
        byteorder = sys.byteorder
    bytes = get_byte_view(rgba_image)
    assert bytes.shape[-1] == 4
    if byteorder == 'little':
        return bytes[..., 0]  # strip A off BGRA
    else:
        return bytes[..., 3]  # strip A off ARGB


def get_alpha_view(rgba_image, byteorder=None):
    if byteorder is None:
        byteorder = sys.byteorder
    bytes = get_byte_view(rgba_image)
    assert bytes.shape[-1] == 4
    if byteorder == 'little':
        return bytes[..., 3]  # A of BGRA
    else:
        return bytes[..., 0]  # A of ARGB


class DrawingContext:
    """
        Path commands (begin_path, close_path, move_to, line_to, etc.) should not be intermixed
        with transform commands (translate, scale, rotate).
    """

    # TODO: stroke_fill
    # TODO: circle

    __image_id = 0
    __image_id_lock = threading.RLock()

    def __init__(self):
        self.commands = []
        self.binary_commands = bytearray()
        self.save_count = 0
        self.images = dict()

    def copy_from(self, drawing_context):
        assert self.save_count == 0
        assert drawing_context.save_count == 0
        self.commands = drawing_context.commands
        self.binary_commands = drawing_context.binary_commands
        self.images = drawing_context.images

    def add(self, drawing_context):
        self.commands.extend(drawing_context.commands)
        self.binary_commands.extend(drawing_context.binary_commands)
        self.images.update(drawing_context.images)

    def clear(self):
        self.commands = []
        self.binary_commands = []
        self.save_count = 0
        self.images = dict()

    def to_js(self):
        js = ""
        for command in self.commands:
            command_id = command[0]
            command_args = command[1:]
            if command_id == "save":
                js += "ctx.save();"
            elif command_id == "restore":
                js += "ctx.restore();"
            elif command_id == "beginPath":
                js += "ctx.beginPath();"
            elif command_id == "closePath":
                js += "ctx.closePath();"
            elif command_id == "clip":
                js += "ctx.beginPath();"
                js += "ctx.rect({0}, {1}, {2}, {3});".format(*command_args)
                js += "ctx.clip();"
            elif command_id == "translate":
                js += "ctx.translate({0}, {1});".format(*command_args)
            elif command_id == "scale":
                js += "ctx.scale({0}, {1});".format(*command_args)
            elif command_id == "rotate":
                js += "ctx.rotate({0});".format(*command_args)
            elif command_id == "moveTo":
                js += "ctx.moveTo({0}, {1});".format(*command_args)
            elif command_id == "lineTo":
                js += "ctx.lineTo({0}, {1});".format(*command_args)
            elif command_id == "rect":
                js += "ctx.rect({0}, {1}, {2}, {3});".format(*command_args)
            elif command_id == "arc":
                x, y, r, sa, ea, ac = command_args
                js += "ctx.arc({0}, {1}, {2}, {3}, {4}, {5});".format(x, y, r, sa, ea, "true" if ac else "false")
            elif command_id == "arcTo":
                x1, y1, x2, y2, r = command_args
                js += "ctx.arcTo({0}, {1}, {2}, {3}, {4});".format(x1, y1, x2, y2, r)
            elif command_id == "cubicTo":
                x1, y1, x2, y2, x, y = command_args
                js += "ctx.bezierCurveTo({0}, {1}, {2}, {3}, {4}, {5});".format(x1, y1, x2, y2, x, y)
            elif command_id == "quadraticTo":
                x1, y1, x, y = command_args
                js += "ctx.quadraticCurveTo({0}, {1}, {2}, {3});".format(x1, y1, x, y)
            elif command_id == "image":
                w, h, image, image_id, a, b, c, d = command_args
                js += "ctx.rect({0}, {1}, {2}, {3});".format(a, b, c, d)
            elif command_id == "data":
                w, h, data, data_id, a, b, c, d, low, high, color_table = command_args
                js += "ctx.rect({0}, {1}, {2}, {3});".format(a, b, c, d)
            elif command_id == "stroke":
                js += "ctx.stroke();"
            elif command_id == "sleep":
                pass  # used for performance testing
            elif command_id == "fill":
                js += "ctx.fill();"
            elif command_id == "fillText":
                text, x, y, max_width = command_args
                js += "ctx.fillText('{0}', {1}, {2}{3});".format(xml.sax.saxutils.escape(text), x, y, ", {0}".format(max_width) if max_width else "")
            elif command_id == "fillStyleGradient":
                command_var = command_args[0]
                js += "ctx.fillStyle = {0};".format("grad" + str(command_var))
            elif command_id == "fillStyle":
                js += "ctx.fillStyle = '{0}';".format(*command_args)
            elif command_id == "font":
                js += "ctx.font = '{0}';".format(*command_args)
            elif command_id == "textAlign":
                js += "ctx.textAlign = '{0}';".format(*command_args)
            elif command_id == "textBaseline":
                js += "ctx.textBaseline = '{0}';".format(*command_args)
            elif command_id == "strokeStyle":
                js += "ctx.strokeStyle = '{0}';".format(*command_args)
            elif command_id == "lineWidth":
                js += "ctx.lineWidth = {0};".format(*command_args)
            elif command_id == "lineDash":
                js += "ctx.lineDash = {0};".format(*command_args)
            elif command_id == "lineCap":
                js += "ctx.lineCap = '{0}';".format(*command_args)
            elif command_id == "lineJoin":
                js += "ctx.lineJoin = '{0}';".format(*command_args)
            elif command_id == "gradient":
                command_var, width, height, x1, y1, x2, y2 = command_args  # pylint: disable=invalid-name
                js_var = "grad" + str(command_var)
                js += "var {0} = ctx.createLinearGradient({1}, {2}, {3}, {4});".format(js_var, x1, y1, x2 - x1, y2 - y1)
            elif command_id == "colorStop":
                command_var, x, color = command_args
                js_var = "grad" + str(command_var)
                js += "{0}.addColorStop({1}, '{2}');".format(js_var, x, color)
        return js

    def to_svg(self, size, viewbox):
        svg = ""
        defs = ""
        path = ""
        next_clip_id = 1
        transform = list()
        closers = list()
        fill_style = None
        fill_opacity = 1.0
        stroke_style = None
        stroke_opacity = 1.0
        line_cap = "square"
        line_join = "bevel"
        line_width = 1.0
        line_dash = None
        text_anchor = "start"
        text_baseline = "alphabetic"
        font_style = None
        font_weight = None
        font_size = None
        font_unit = None
        font_family = None
        contexts = collections.deque()
        gradient_start = None
        gradient_stops = list()

        # make a SVG 1.1 compatible color, opacity tuple
        def parse_color(color_str: str) -> typing.Tuple[str, float]:
            color_str = ''.join(color_str.split())
            if color_str.startswith("rgba"):
                c = re.split("rgba\((\d+),(\d+),(\d+),([\d.]+)\)", color_str)
                return f"rgb({c[1]}, {c[2]}, {c[3]})", float(c[4])
            return color_str, 1.0

        for command in self.commands:
            command_id = command[0]
            #logging.debug(command_id)
            command_args = command[1:]
            if command_id == "save":
                context = dict()
                context["path"] = path
                context["transform"] = copy.deepcopy(transform)
                context["fill_style"] = fill_style
                context["fill_opacity"] = fill_opacity
                context["stroke_style"] = stroke_style
                context["stroke_opacity"] = stroke_opacity
                context["line_cap"] = line_cap
                context["line_join"] = line_join
                context["line_width"] = line_width
                context["line_dash"] = line_dash
                context["font_style"] = font_style
                context["font_weight"] = font_weight
                context["font_size"] = font_size
                context["font_unit"] = font_unit
                context["font_family"] = font_family
                context["text_anchor"] = text_anchor
                context["text_baseline"] = text_baseline
                context["closers"] = copy.deepcopy(closers)
                closers = list()
                contexts.append(context)
            elif command_id == "restore":
                svg += "".join(closers)
                context = contexts.pop()
                path = context["path"]
                transform = context["transform"]
                fill_style = context["fill_style"]
                fill_opacity = context["fill_opacity"]
                font_style = context["font_style"]
                font_weight = context["font_weight"]
                font_size = context["font_size"]
                font_unit = context["font_unit"]
                font_family = context["font_family"]
                text_anchor = context["text_anchor"]
                text_baseline = context["text_baseline"]
                stroke_style = context["stroke_style"]
                stroke_opacity = context["stroke_opacity"]
                line_cap = context["line_cap"]
                line_join = context["line_join"]
                line_width = context["line_width"]
                line_dash = context["line_dash"]
                closers = context["closers"]
            elif command_id == "beginPath":
                path = ""
            elif command_id == "closePath":
                path += " Z"
            elif command_id == "moveTo":
                path += " M {0} {1}".format(*command_args)
            elif command_id == "lineTo":
                path += " L {0} {1}".format(*command_args)
            elif command_id == "rect":
                x, y, w, h = command_args
                path += " M {0} {1}".format(x, y)
                path += " L {0} {1}".format(x + w, y)
                path += " L {0} {1}".format(x + w, y + h)
                path += " L {0} {1}".format(x, y + h)
                path += " Z"
            elif command_id == "arc":
                x, y, r, sa, ea, ac = command_args
                # js += "ctx.arc({0}, {1}, {2}, {3}, {4}, {5});".format(x, y, r, sa, ea, "true" if ac else "false")
            elif command_id == "arcTo":
                x1, y1, x2, y2, r = command_args
                # js += "ctx.arcTo({0}, {1}, {2}, {3}, {4});".format(x1, y1, x2, y2, r)
            elif command_id == "cubicTo":
                path += " C {0} {1}, {2} {3}, {4} {5}".format(*command_args)
            elif command_id == "quadraticTo":
                path += " Q {0} {1}, {2} {3}".format(*command_args)
            elif command_id == "clip":
                x, y, w, h = command_args
                clip_id = "clip" + str(next_clip_id)
                next_clip_id += 1
                transform_str = " transform='{0}'".format(" ".join(transform)) if len(transform) > 0 else ""
                defs_format_str = "<clipPath id='{0}'><rect x='{1}' y='{2}' width='{3}' height='{4}'{5} /></clipPath>"
                defs += defs_format_str.format(clip_id, x, y, w, h, transform_str)
                svg += "<g style='clip-path: url(#{0});'>".format(clip_id)
                closers.append("</g>")
            elif command_id == "translate":
                transform.append("translate({0},{1})".format(*command_args))
            elif command_id == "scale":
                transform.append("scale({0},{1})".format(*command_args))
            elif command_id == "rotate":
                transform.append("rotate({0})".format(*command_args))
            elif command_id == "image":
                w, h, image, image_id, a, b, c, d = command_args
                png_file = io.BytesIO()
                rgba_data = get_rgba_view_from_rgba_data(image)
                # image compression is time consuming. pass parameters to make this step as fast as possible.
                # see nionswift-642.
                imageio.imwrite(png_file, rgba_data[..., (2,1,0,3)], "png", optimize=False, compress_level=1)
                png_encoded = base64.b64encode(png_file.getvalue()).decode('utf=8')
                transform_str = " transform='{0}'".format(" ".join(transform)) if len(transform) > 0 else ""
                svg_format_str = "<image x='{0}' y='{1}' width='{2}' height='{3}' xlink:href='data:image/png;base64,{4}'{5} />"
                svg += svg_format_str.format(a, b, c, d, png_encoded, transform_str)
            elif command_id == "data":
                w, h, data, data_id, a, b, c, d, low, high, color_table, color_table_image_id = command_args
                m = 255.0 / (high - low) if high != low else 1
                image = numpy.empty(data.shape, numpy.uint32)
                if color_table is not None:
                    adj_color_table = numpy.empty(color_table.shape, numpy.uint32)
                    # ordering of color_table is BGRA
                    # ordering of adj_color_table is RGBA
                    get_byte_view(adj_color_table)[:, 0] = get_byte_view(color_table)[:, 2]
                    get_byte_view(adj_color_table)[:, 1] = get_byte_view(color_table)[:, 1]
                    get_byte_view(adj_color_table)[:, 2] = get_byte_view(color_table)[:, 0]
                    get_byte_view(adj_color_table)[:, 3] = get_byte_view(color_table)[:, 3]
                    clipped_array = numpy.clip((m * (data - low)).astype(int), 0, 255).astype(numpy.uint8)
                    image[:] = adj_color_table[clipped_array]
                else:
                    clipped_array = numpy.clip(data, low, high)
                    numpy.subtract(clipped_array, low, out=clipped_array)
                    numpy.multiply(clipped_array, m, out=clipped_array)
                    get_red_view(image)[:] = clipped_array
                    get_green_view(image)[:] = clipped_array
                    get_blue_view(image)[:] = clipped_array
                    get_alpha_view(image)[:] = 255
                png_file = io.BytesIO()
                # image compression is time consuming. pass parameters to make this step as fast as possible.
                # see nionswift-642.
                imageio.imwrite(png_file, get_rgba_view_from_rgba_data(image), "png", optimize=False, compress_level=1)
                png_encoded = base64.b64encode(png_file.getvalue()).decode('utf=8')
                transform_str = " transform='{0}'".format(" ".join(transform)) if len(transform) > 0 else ""
                svg_format_str = "<image x='{0}' y='{1}' width='{2}' height='{3}' xlink:href='data:image/png;base64,{4}'{5} />"
                svg += svg_format_str.format(a, b, c, d, png_encoded, transform_str)
            elif command_id == "stroke":
                if stroke_style is not None:
                    transform_str = " transform='{0}'".format(" ".join(transform)) if len(transform) > 0 else ""
                    dash_str = " stroke-dasharray='{0}, {1}'".format(line_dash, line_dash) if line_dash else ""
                    svg += f"<path d='{path}' fill='none' stroke='{stroke_style}' stroke-opacity='{stroke_opacity}' stroke-width='{line_width}' stroke-linejoin='{line_join}' stroke-linecap='{line_cap}'{dash_str}{transform_str} />"
            elif command_id == "sleep":
                pass  # used for performance testing
            elif command_id == "fill":
                if fill_style is not None:
                    transform_str = " transform='{0}'".format(" ".join(transform)) if len(transform) > 0 else ""
                    svg += f"<path d='{path}' fill='{fill_style}' fill-opacity='{fill_opacity}' stroke='none'{transform_str} />"
            elif command_id == "fillText":
                text, x, y, max_width = command_args
                transform_str = " transform='{0}'".format(" ".join(transform)) if len(transform) > 0 else ""
                font_str = ""
                if font_style:
                    font_str += " font-style='{0}'".format(font_style)
                if font_weight:
                    font_str += " font-weight='{0}'".format(font_weight)
                if font_size:
                    font_str += " font-size='{0}{1}'".format(font_size, font_unit)
                if font_family:
                    font_str += " font-family='{0}'".format(font_family)
                if fill_style:
                    font_str += " fill='{0}'".format(fill_style)
                if fill_opacity < 1.0:
                    font_str += " fill-opacity='{0}'".format(fill_opacity)
                svg_format_str = "<text x='{0}' y='{1}' text-anchor='{3}' alignment-baseline='{4}'{5}{6}>{2}</text>"
                svg += svg_format_str.format(x, y, xml.sax.saxutils.escape(text), text_anchor, text_baseline, font_str,
                                             transform_str)
            elif command_id == "fillStyleGradient":
                command_var = command_args[0]
                defs += gradient_start + "".join(gradient_stops) + "</linearGradient>"
                fill_style = "url(#{0})".format("grad" + str(command_var))
            elif command_id == "fillStyle":
                fill_style, fill_opacity = parse_color(command_args[0])
            elif command_id == "font":
                font_style = None
                font_weight = None
                font_size = None
                font_unit = None
                font_family = None
                for font_part in [s for s in command_args[0].split(" ") if s]:
                    if font_part == "italic":
                        font_style = "italic"
                    elif font_part == "bold":
                        font_weight = "bold"
                    elif font_part.endswith("px") and int(font_part[0:-2]) > 0:
                        font_size = int(font_part[0:-2])
                        font_unit = "px"
                    elif font_part.endswith("pt") and int(font_part[0:-2]) > 0:
                        font_size = int(font_part[0:-2])
                        font_unit = "pt"
                    else:
                        font_family = font_part
            elif command_id == "textAlign":
                text_anchors = {"start": "start", "end": "end", "left": "start", "center": "middle", "right": "end"}
                text_anchor = text_anchors.get(command_args[0], "start")
            elif command_id == "textBaseline":
                text_baselines = {"top": "hanging", "hanging": "hanging", "middle": "middle",
                                  "alphabetic": "alphabetic", "ideaographic": "ideaographic", "bottom": "bottom"}
                text_baseline = text_baselines.get(command_args[0], "alphabetic")
            elif command_id == "strokeStyle":
                stroke_style, stroke_opacity = parse_color(command_args[0])
            elif command_id == "lineWidth":
                line_width = command_args[0]
            elif command_id == "lineDash":
                line_dash = command_args[0]
            elif command_id == "lineCap":
                line_caps = {"square": "square", "round": "round", "butt": "butt"}
                line_cap = line_caps.get(command_args[0], "square")
            elif command_id == "lineJoin":
                line_joins = {"round": "round", "miter": "miter", "bevel": "bevel"}
                line_join = line_joins.get(command_args[0], "bevel")
            elif command_id == "gradient":
                # assumes that gradient will be used immediately after being
                # declared and stops being defined. this is currently enforced by
                # the way the commands are generated in drawing context.
                command_var, w, h, x1, y1, x2, y2 = command_args
                grad_id = "grad" + str(command_var)
                gradient_start = "<linearGradient id='{0}' x1='{1}' y1='{2}' x2='{3}' y2='{4}'>".format(grad_id,
                                                                                                        float(x1 / w),
                                                                                                        float(y1 / h),
                                                                                                        float(x2 / w),
                                                                                                        float(y2 / h))
            elif command_id == "colorStop":
                command_var, x, color = command_args
                gradient_stops.append("<stop offset='{0}%' stop-color='{1}' />".format(int(x * 100), color))
            else:
                logging.debug("Unknown command %s", command)
        xmlns = "xmlns='http://www.w3.org/2000/svg' xmlns:xlink='http://www.w3.org/1999/xlink'"
        viewbox_str = "{0} {1} {2} {3}".format(viewbox.left, viewbox.top, viewbox.width, viewbox.height)
        result = "<svg version='1.1' baseProfile='full' width='{0}' height='{1}' viewBox='{2}' {3}>".format(size.width,
                                                                                                            size.height,
                                                                                                            viewbox_str,
                                                                                                            xmlns)
        result += "<defs>" + defs + "</defs>"
        result += svg
        result += "</svg>"
        return result

    @contextmanager
    def saver(self):
        self.save()
        try:
            yield
        finally:
           self.restore()

    def save(self):
        self.commands.append(("save", ))
        self.binary_commands.extend(b"save")
        self.save_count += 1

    def restore(self):
        self.commands.append(("restore", ))
        self.binary_commands.extend(b"rest")
        self.save_count -= 1

    def begin_layer(self, layer_id: int, layer_seed: int, a, b, c, d) -> None:
        self.commands.append(("begin_layer", int(layer_id), int(layer_seed), float(a), float(b), float(c), float(d)))
        self.binary_commands.extend(struct.pack("4siiffff", b"bgly", int(layer_id), int(layer_seed), float(a), float(b), float(c), float(d)))

    def end_layer(self, layer_id: int, layer_seed: int, a, b, c, d) -> None:
        self.commands.append(("end_layer", int(layer_id), int(layer_seed), float(a), float(b), float(c), float(d)))
        self.binary_commands.extend(struct.pack("4siiffff", b"enly", int(layer_id), int(layer_seed), float(a), float(b), float(c), float(d)))

    def begin_path(self):
        self.commands.append(("beginPath", ))
        self.binary_commands.extend(b"bpth")

    def close_path(self):
        self.commands.append(("closePath", ))
        self.binary_commands.extend(b"cpth")

    def clip_rect(self, a, b, c, d):
        self.commands.append(("clip", float(a), float(b), float(c), float(d)))
        self.binary_commands.extend(struct.pack("4sffff", b"clip", float(a), float(b), float(c), float(d)))

    def translate(self, x, y):
        self.commands.append(("translate", float(x), float(y)))
        self.binary_commands.extend(struct.pack("4sff", b"tran", float(x), float(y)))

    def scale(self, x, y):
        self.commands.append(("scale", float(x), float(y)))
        self.binary_commands.extend(struct.pack("4sff", b"scal", float(x), float(y)))

    def rotate(self, radians):
        self.commands.append(("rotate", math.degrees(float(radians))))
        self.binary_commands.extend(struct.pack("4sf", b"rota", math.degrees(float(radians))))

    def move_to(self, x, y):
        self.commands.append(("moveTo", float(x), float(y)))
        self.binary_commands.extend(struct.pack("4sff", b"move", float(x), float(y)))

    def line_to(self, x, y):
        self.commands.append(("lineTo", float(x), float(y)))
        self.binary_commands.extend(struct.pack("4sff", b"line", float(x), float(y)))

    def rect(self, l, t, w, h):
        self.commands.append(("rect", float(l), float(t), float(w), float(h)))
        self.binary_commands.extend(struct.pack("4sffff", b"rect", float(l), float(t), float(w), float(h)))

    def round_rect(self, x, y, w, h, r):
        self.move_to(x + r, y)
        self.arc_to(x + w, y, x + w, y + r, r)
        self.arc_to(x + w, y + h, x + w - r, y + h, r)
        self.arc_to(x, y + h, x, y + h - r, r)
        self.arc_to(x, y, x + r, y, r)
        self.close_path()

    def arc(self, x, y, r, sa, ea, ac=False):
        self.commands.append(("arc", float(x), float(y), float(r), float(sa), float(ea), bool(ac)))
        self.binary_commands.extend(struct.pack("4sfffffi", b"arc ", float(x), float(y), float(r), float(sa), float(ea), bool(ac)))

    def arc_to(self, x1, y1, x2, y2, r):
        self.commands.append(("arcTo", float(x1), float(y1), float(x2), float(y2), float(r)))
        self.binary_commands.extend(struct.pack("4sfffff", b"arct", float(x1), float(y1), float(x2), float(y2), float(r)))

    def bezier_curve_to(self, x1, y1, x2, y2, x, y):
        self.commands.append(("cubicTo", float(x1), float(y1), float(x2), float(y2), float(x), float(y)))
        self.binary_commands.extend(struct.pack("4sffffff", b"cubc", float(x1), float(y1), float(x2), float(y2), float(x), float(y)))

    def quadratic_curve_to(self, x1, y1, x, y):
        self.commands.append(("quadraticTo", float(x1), float(y1), float(x), float(y)))
        self.binary_commands.extend(struct.pack("4sffff", b"quad", float(x1), float(y1), float(x), float(y)))

    def draw_image(self, img, x, y, width, height):
        # img should be rgba pack, uint32
        assert img.dtype == numpy.uint32
        with DrawingContext.__image_id_lock:
            DrawingContext.__image_id += 1
            image_id = DrawingContext.__image_id
        self.commands.append(
            ("image", img.shape[1], img.shape[0], img, int(image_id), float(x), float(y), float(width), float(height)))
        self.images[str(image_id)] = img
        self.binary_commands.extend(struct.pack("4siiiffff", b"imag", img.shape[1], img.shape[0], int(image_id), float(x), float(y), float(width), float(height)))

    def draw_data(self, img, x, y, width, height, low, high, color_map_data):
        # img should be float
        assert img.dtype == numpy.float32
        with DrawingContext.__image_id_lock:
            DrawingContext.__image_id += 1
            image_id = DrawingContext.__image_id
            if color_map_data is not None:
                DrawingContext.__image_id += 1
                color_map_image_id = DrawingContext.__image_id
            else:
                color_map_image_id = 0
        self.images[str(image_id)] = img
        if color_map_data is not None:
            self.images[str(color_map_image_id)] = color_map_data
        self.commands.append(
            ("data", img.shape[1], img.shape[0], img, int(image_id), float(x), float(y), float(width), float(height), float(low), float(high), color_map_data, int(color_map_image_id)))
        self.binary_commands.extend(struct.pack("4siiiffffffi", b"data", img.shape[1], img.shape[0], int(image_id), float(x), float(y), float(width), float(height), float(low), float(high), int(color_map_image_id)))

    def stroke(self):
        self.commands.append(("stroke", ))
        self.binary_commands.extend(b"strk")

    def sleep(self, duration):
        self.commands.append(("sleep", float(duration)))
        self.binary_commands.extend(struct.pack("4sf", b"slep", float(duration)))

    def mark_latency(self):
        self.commands.append(("latency", time.perf_counter()))
        self.binary_commands.extend(struct.pack("<4sd", b"latn", time.perf_counter()))

    def message(self, text):
        self.commands.append(("message", text))
        text_encoded = text.encode("utf-8")
        self.binary_commands.extend(struct.pack("4si{}s0i".format(len(text_encoded)), b"mesg", len(text_encoded), text_encoded))

    def timestamp(self, timestamp):
        self.commands.append(("timestamp", timestamp))
        timestamp_encoded = timestamp.encode("utf-8")
        self.binary_commands.extend(struct.pack("4si{}s0i".format(len(timestamp_encoded)), b"time", len(timestamp_encoded), timestamp_encoded))

    def fill(self):
        self.commands.append(("fill", ))
        self.binary_commands.extend(b"fill")

    def fill_text(self, text, x, y, max_width=None):
        text = str(text) if text is not None else str()
        self.commands.append(("fillText", text, float(x), float(y), float(max_width) if max_width else 0))
        text_encoded = text.encode("utf-8")
        self.binary_commands.extend(struct.pack("4si{}sfff".format(len(text_encoded)), b"text", len(text_encoded), text_encoded, float(x), float(y), float(max_width) if max_width else 0))

    @property
    def fill_style(self):
        raise NotImplementedError()

    @fill_style.setter
    def fill_style(self, a):
        a = a or "rgba(0, 0, 0, 0.0)"
        if isinstance(a, DrawingContext.LinearGradient):
            self.commands.extend(a.commands)
            self.commands.append(("fillStyleGradient", int(a.command_var)))
            self.binary_commands.extend(a.binary_commands)
            self.binary_commands.extend(struct.pack("4si", b"flsg", int(a.command_var)))
        else:
            self.commands.append(("fillStyle", str(a)))
            a_encoded = a.encode("utf-8")
            self.binary_commands.extend(struct.pack("4si{}s0i".format(len(a_encoded)), b"flst", len(a_encoded), a_encoded))

    @property
    def font(self):
        raise NotImplementedError()

    @font.setter
    def font(self, a):
        """
            Set the text font.

            Supports 'normal', 'bold', 'italic', size specific as '14px', and font-family.
        """
        self.commands.append(("font", str(a)))
        a_encoded = a.encode("utf-8")
        self.binary_commands.extend(struct.pack("4si{}s0i".format(len(a_encoded)), b"font", len(a_encoded), a_encoded))

    def __get_text_align(self):
        raise NotImplementedError()

    def __set_text_align(self, a):
        """
            Set text alignment.

            Valid values are 'start', 'end', 'left', 'center', 'right'. Default is 'start'.

            Default is 'start'.
        """
        self.commands.append(("textAlign", str(a)))
        a_encoded = a.encode("utf-8")
        self.binary_commands.extend(struct.pack("4si{}s0i".format(len(a_encoded)), b"algn", len(a_encoded), a_encoded))

    text_align = property(__get_text_align, __set_text_align)

    def __get_text_baseline(self):
        raise NotImplementedError()

    def __set_text_baseline(self, a):
        """
            Set the text baseline.

            Valid values are 'top', 'hanging', 'middle', 'alphabetic', 'ideographic', and 'bottom'.

            Default is 'alphabetic'.
        """
        self.commands.append(("textBaseline", str(a)))
        a_encoded = a.encode("utf-8")
        self.binary_commands.extend(struct.pack("4si{}s0i".format(len(a_encoded)), b"tbas", len(a_encoded), a_encoded))

    text_baseline = property(__get_text_baseline, __set_text_baseline)

    def __get_stroke_style(self):
        raise NotImplementedError()

    def __set_stroke_style(self, a):
        a = a or "rgba(0, 0, 0, 0.0)"
        self.commands.append(("strokeStyle", str(a)))
        a_encoded = a.encode("utf-8")
        self.binary_commands.extend(struct.pack("4si{}s0i".format(len(a_encoded)), b"stst", len(a_encoded), a_encoded))

    stroke_style = property(__get_stroke_style, __set_stroke_style)

    def __get_line_width(self):
        raise NotImplementedError()

    def __set_line_width(self, a):
        self.commands.append(("lineWidth", float(a)))
        self.binary_commands.extend(struct.pack("4sf", b"linw", float(a)))

    line_width = property(__get_line_width, __set_line_width)

    def __get_line_dash(self):
        raise NotImplementedError()

    def __set_line_dash(self, a):
        """ Set the line dash. Takes a single value with the length of the dash. """
        self.commands.append(("lineDash", float(a)))
        self.binary_commands.extend(struct.pack("4sf", b"ldsh", float(a)))

    line_dash = property(__get_line_dash, __set_line_dash)

    def __get_line_cap(self):
        raise NotImplementedError()

    def __set_line_cap(self, a):
        """ Set the line join. Valid values are 'square', 'round', 'butt'. Default is 'square'. """
        self.commands.append(("lineCap", str(a)))
        a_encoded = a.encode("utf-8")
        self.binary_commands.extend(struct.pack("4si{}s0i".format(len(a_encoded)), b"lcap", len(a_encoded), a_encoded))

    line_cap = property(__get_line_cap, __set_line_cap)

    def __get_line_join(self):
        raise NotImplementedError()

    def __set_line_join(self, a):
        """ Set the line join. Valid values are 'round', 'miter', 'bevel'. Default is 'bevel'. """
        self.commands.append(("lineJoin", str(a)))
        a_encoded = a.encode("utf-8")
        self.binary_commands.extend(struct.pack("4si{}s0i".format(len(a_encoded)), b"lnjn", len(a_encoded), a_encoded))

    line_join = property(__get_line_join, __set_line_join)

    class LinearGradient:
        next = 1

        def __init__(self, width, height, x1, y1, x2, y2):  # pylint: disable=invalid-name
            self.commands = []
            self.binary_commands = []
            self.command_var = DrawingContext.LinearGradient.next
            self.commands.append(("gradient", self.command_var, float(width), float(height), float(x1), float(y1), float(x2), float(y2)))
            self.binary_commands.extend(struct.pack("4siffffff", b"grad", self.command_var, float(width), float(height), float(x1), float(y1), float(x2), float(y2)))
            DrawingContext.LinearGradient.next += 1

        def add_color_stop(self, x, color):
            self.commands.append(("colorStop", self.command_var, float(x), str(color)))
            color_encoded = color.encode("utf-8")
            self.binary_commands.extend(struct.pack("4sifi{}s0i".format(len(color_encoded)), b"grcs", self.command_var, float(x), len(color_encoded), color_encoded))

    def create_linear_gradient(self, width, height, x1, y1, x2, y2):  # pylint: disable=invalid-name
        gradient = DrawingContext.LinearGradient(width, height, x1, y1, x2, y2)
        return gradient

    def statistics(self, stat_id):
        self.commands.append(("statistics", str(stat_id)))
        stat_id_encoded = stat_id.encode("utf-8")
        self.binary_commands.extend(struct.pack("4si{}s0i".format(len(stat_id_encoded)), b"stat", len(stat_id_encoded), stat_id_encoded))

# https://www.w3.org/TR/SVG11/types.html#ColorKeywords

# processed with:

"""
import re

with open("colors.txt", "r") as f:
    while True:
        color_line = f.readline()
        if not color_line:
            break
        color_line = color_line.strip()
        rgb_line = f.readline().strip().replace(" ", "")
        c = re.split("rgb\((\d+),(\d+),(\d+)\)", rgb_line)
        print(f"\t\"{color_line}\": \"#{int(c[1]):02x}{int(c[2]):02x}{int(c[3]):02x}\",")
"""

svg_color_map = {
    "aliceblue": "#f0f8ff",
    "antiquewhite": "#faebd7",
    "aqua": "#00ffff",
    "aquamarine": "#7fffd4",
    "azure": "#f0ffff",
    "beige": "#f5f5dc",
    "bisque": "#ffe4c4",
    "black": "#000000",
    "blanchedalmond": "#ffebcd",
    "blue": "#0000ff",
    "blueviolet": "#8a2be2",
    "brown": "#a52a2a",
    "burlywood": "#deb887",
    "cadetblue": "#5f9ea0",
    "chartreuse": "#7fff00",
    "chocolate": "#d2691e",
    "coral": "#ff7f50",
    "cornflowerblue": "#6495ed",
    "cornsilk": "#fff8dc",
    "crimson": "#dc143c",
    "cyan": "#00ffff",
    "darkblue": "#00008b",
    "darkcyan": "#008b8b",
    "darkgoldenrod": "#b8860b",
    "darkgray": "#a9a9a9",
    "darkgreen": "#006400",
    "darkgrey": "#a9a9a9",
    "darkkhaki": "#bdb76b",
    "darkmagenta": "#8b008b",
    "darkolivegreen": "#556b2f",
    "darkorange": "#ff8c00",
    "darkorchid": "#9932cc",
    "darkred": "#8b0000",
    "darksalmon": "#e9967a",
    "darkseagreen": "#8fbc8f",
    "darkslateblue": "#483d8b",
    "darkslategray": "#2f4f4f",
    "darkslategrey": "#2f4f4f",
    "darkturquoise": "#00ced1",
    "darkviolet": "#9400d3",
    "deeppink": "#ff1493",
    "deepskyblue": "#00bfff",
    "dimgray": "#696969",
    "dimgrey": "#696969",
    "dodgerblue": "#1e90ff",
    "firebrick": "#b22222",
    "floralwhite": "#fffaf0",
    "forestgreen": "#228b22",
    "fuchsia": "#ff00ff",
    "gainsboro": "#dcdcdc",
    "ghostwhite": "#f8f8ff",
    "gold": "#ffd700",
    "goldenrod": "#daa520",
    "gray": "#808080",
    "grey": "#808080",
    "green": "#008000",
    "greenyellow": "#adff2f",
    "honeydew": "#f0fff0",
    "hotpink": "#ff69b4",
    "indianred": "#cd5c5c",
    "indigo": "#4b0082",
    "ivory": "#fffff0",
    "khaki": "#f0e68c",
    "lavender": "#e6e6fa",
    "lavenderblush": "#fff0f5",
    "lawngreen": "#7cfc00",
    "lemonchiffon": "#fffacd",
    "lightblue": "#add8e6",
    "lightcoral": "#f08080",
    "lightcyan": "#e0ffff",
    "lightgoldenrodyellow": "#fafad2",
    "lightgray": "#d3d3d3",
    "lightgreen": "#90ee90",
    "lightgrey": "#d3d3d3",
    "lightpink": "#ffb6c1",
    "lightsalmon": "#ffa07a",
    "lightseagreen": "#20b2aa",
    "lightskyblue": "#87cefa",
    "lightslategray": "#778899",
    "lightslategrey": "#778899",
    "lightsteelblue": "#b0c4de",
    "lightyellow": "#ffffe0",
    "lime": "#00ff00",
    "limegreen": "#32cd32",
    "linen": "#faf0e6",
    "magenta": "#ff00ff",
    "maroon": "#800000",
    "mediumaquamarine": "#66cdaa",
    "mediumblue": "#0000cd",
    "mediumorchid": "#ba55d3",
    "mediumpurple": "#9370db",
    "mediumseagreen": "#3cb371",
    "mediumslateblue": "#7b68ee",
    "mediumspringgreen": "#00fa9a",
    "mediumturquoise": "#48d1cc",
    "mediumvioletred": "#c71585",
    "midnightblue": "#191970",
    "mintcream": "#f5fffa",
    "mistyrose": "#ffe4e1",
    "moccasin": "#ffe4b5",
    "navajowhite": "#ffdead",
    "navy": "#000080",
    "oldlace": "#fdf5e6",
    "olive": "#808000",
    "olivedrab": "#6b8e23",
    "orange": "#ffa500",
    "orangered": "#ff4500",
    "orchid": "#da70d6",
    "palegoldenrod": "#eee8aa",
    "palegreen": "#98fb98",
    "paleturquoise": "#afeeee",
    "palevioletred": "#db7093",
    "papayawhip": "#ffefd5",
    "peachpuff": "#ffdab9",
    "peru": "#cd853f",
    "pink": "#ffc0cb",
    "plum": "#dda0dd",
    "powderblue": "#b0e0e6",
    "purple": "#800080",
    "red": "#ff0000",
    "rosybrown": "#bc8f8f",
    "royalblue": "#4169e1",
    "saddlebrown": "#8b4513",
    "salmon": "#fa8072",
    "sandybrown": "#f4a460",
    "seagreen": "#2e8b57",
    "seashell": "#fff5ee",
    "sienna": "#a0522d",
    "silver": "#c0c0c0",
    "skyblue": "#87ceeb",
    "slateblue": "#6a5acd",
    "slategray": "#708090",
    "slategrey": "#708090",
    "snow": "#fffafa",
    "springgreen": "#00ff7f",
    "steelblue": "#4682b4",
    "tan": "#d2b48c",
    "teal": "#008080",
    "thistle": "#d8bfd8",
    "tomato": "#ff6347",
    "turquoise": "#40e0d0",
    "violet": "#ee82ee",
    "wheat": "#f5deb3",
    "white": "#ffffff",
    "whitesmoke": "#f5f5f5",
    "yellow": "#ffff00",
    "yellowgreen": "#9acd32",
}


svg_color_reverse_map = {v: k for k, v in svg_color_map.items()}


def color_without_alpha(color: typing.Optional[str]) -> typing.Optional[str]:
    if not color:
        return None
    color = color.strip().replace(" ", "")
    c = re.split("rgba\((\d+),(\d+),(\d+),([\d.]+)\)", color)
    if len(c) > 1:
        return f"#{int(c[1]):02x}{int(c[2]):02x}{int(c[3]):02x}"
    c = re.split("rgb\((\d+),(\d+),(\d+)\)", color)
    if len(c) > 1:
        return f"#{int(c[1]):02x}{int(c[2]):02x}{int(c[3]):02x}"
    if color.startswith("#"):
        if len(color) == 9:
            return f"#{color.lower()[3:]}"
        if len(color) == 7:
            return color.lower()
    return color


def named_color_without_alpha(color: str) -> typing.Optional[str]:
    color_with_alpha = color_without_alpha(color)
    return svg_color_reverse_map.get(color_with_alpha, color_with_alpha) if color_with_alpha else None


def hex_color(color: str) -> typing.Optional[str]:
    if not color:
        return None
    color = color.strip().replace(" ", "")
    c = re.split("rgba\((\d+),(\d+),(\d+),([\d.]+)\)", color)
    if len(c) > 1:
        return f"#{int(255 * float(c[4])):02x}{int(c[1]):02x}{int(c[2]):02x}{int(c[3]):02x}"
    c = re.split("rgb\((\d+),(\d+),(\d+)\)", color)
    if len(c) > 1:
        return f"#{int(c[1]):02x}{int(c[2]):02x}{int(c[3]):02x}"
    if color.startswith("#"):
        if len(color) == 9 or len(color) == 7:
            return color.lower()
    return svg_color_map.get(color, color)



import asyncio
import functools
import sys
import typing

from nion.ui import CanvasItem
from nion.utils import Geometry


class TrackingCanvasItem(CanvasItem.CanvasItemComposition):

    """Track changes to the mouse without showing the cursor.

    Callers should configure this item by adding a layout and canvas items.

    Callers should listen to the on_mouse_position_changed_by(dx, dy) message.

    Callers should call track(ui, global_pos, window_size) to track the mouse.

    It is an error to call track(...) while already tracking.
    """

    def __init__(self):
        super(TrackingCanvasItem, self).__init__()
        self.wants_mouse_events = True
        self.focusable = True
        self.on_close = None
        self.on_mouse_position_changed_by = None

    def close(self):
        self.on_close = None
        self.on_mouse_position_changed_by = None
        super(TrackingCanvasItem, self).close()

    def mouse_clicked(self, x, y, modifiers):
        self.on_close()
        return True

    def mouse_double_clicked(self, x, y, modifiers):
        self.on_close()
        return True

    def context_menu_event(self, x, y, gx, gy):
        self.on_close()
        return True

    def key_pressed(self, key):
        self.on_close()
        return True

    def grab_mouse(self, gx, gy):
        self.__discard_first = True
        self.root_container.grab_mouse(self, gx, gy)

    def release_mouse(self):
        self.root_container.release_mouse()

    def grabbed_mouse_position_changed(self, dx, dy, modifiers):
        if not self.__discard_first:
            self.on_mouse_position_changed_by(Geometry.IntPoint(x=dx, y=dy))
        self.__discard_first = False
        return True


def start_mouse_tracker(ui, event_loop: asyncio.AbstractEventLoop, canvas_item: CanvasItem.AbstractCanvasItem,
                        mouse_position_changed_by_fn: typing.Callable[[Geometry.IntPoint], None], global_pos: Geometry.IntPoint, size: Geometry.IntSize):

    tracking_canvas_item = TrackingCanvasItem()
    tracking_canvas_item.on_mouse_position_changed_by = mouse_position_changed_by_fn
    tracking_canvas_item.add_canvas_item(canvas_item)

    async def handle_close_later(document_window):
        document_window.request_close()

    def handle_close(document_window):
        tracking_canvas_item.release_mouse()
        event_loop.create_task(handle_close_later(document_window))

    def activation_changed(document_window, activated):
        if not activated:
            handle_close(document_window)

    # create the popup window
    document_window = ui.create_document_window()
    document_window.window_style = "mousegrab"

    def close_window(geometry, state):
        ui.destroy_document_window(document_window)

    document_window.on_about_to_close = close_window

    document_window.on_activation_changed = functools.partial(activation_changed, document_window)
    tracking_canvas_item.on_close = functools.partial(handle_close, document_window)

    # configure canvas widget, attach to document window
    mousegrab_window_pos = global_pos - Geometry.IntPoint(x=size.width, y=size.height / 2)
    document_window.show(size=size, position=mousegrab_window_pos)
    if sys.platform == "win32":
        relative_pos = Geometry.IntPoint()
    else:
        relative_pos = mousegrab_window_pos
        document_window.fill_screen()
    canvas_widget = ui.create_canvas_widget()
    tracking_canvas_item.update_sizing(tracking_canvas_item.sizing.with_fixed_size(size))
    content_row_canvas_item = CanvasItem.CanvasItemComposition()
    content_row_canvas_item.layout = CanvasItem.CanvasItemRowLayout()
    content_row_canvas_item.add_spacing(relative_pos.x)
    content_row_canvas_item.add_canvas_item(tracking_canvas_item)
    content_row_canvas_item.add_stretch()
    content_canvas_item = CanvasItem.CanvasItemComposition()
    content_canvas_item.layout = CanvasItem.CanvasItemColumnLayout()
    content_canvas_item.add_spacing(relative_pos.y)
    content_canvas_item.add_canvas_item(content_row_canvas_item)
    content_canvas_item.add_stretch()
    canvas_widget.canvas_item.add_canvas_item(content_canvas_item)
    document_window.attach(canvas_widget)
    tracking_canvas_item.request_focus()
    tracking_canvas_item.cursor_shape = "blank"
    canvas_widget.set_cursor_shape("blank")
    tracking_canvas_item.grab_mouse(relative_pos.x + size.width // 2, relative_pos.y + size.height // 2)



# standard libraries
import unittest

# third party libraries
# None

# local libraries
from nion.ui import CanvasItem
from nion.ui import ListCanvasItem
from nion.utils import Geometry
from nion.utils import Selection


class ListCanvasItemDelegate:

    @property
    def item_count(self):
        return 4

    def drag_started(self, mouse_index, x, y, modifiers):
        pass


class TestListCanvasItemClass(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_shift_click_extends_selection(self):
        selection = Selection.IndexedSelection()
        delegate = ListCanvasItemDelegate()
        canvas_item = ListCanvasItem.ListCanvasItem(delegate, selection)
        canvas_item.update_layout((0, 0), (320, 100))
        self.assertEqual(selection.indexes, set())
        canvas_item.simulate_click(Geometry.IntPoint(y=120, x=50))
        self.assertEqual(selection.indexes, {1})
        modifiers = CanvasItem.KeyboardModifiers(shift=True)
        canvas_item.simulate_click(Geometry.IntPoint(y=200, x=50), modifiers)
        self.assertEqual(selection.indexes, {1, 2})

    def test_start_drag_does_not_change_selection(self):
        selection = Selection.IndexedSelection()
        delegate = ListCanvasItemDelegate()
        canvas_item = ListCanvasItem.ListCanvasItem(delegate, selection)
        canvas_item.update_layout((0, 0), (320, 100))
        self.assertEqual(selection.indexes, set())
        canvas_item.simulate_drag(Geometry.IntPoint(y=120, x=50), Geometry.IntPoint(y=120, x=500))
        self.assertEqual(selection.indexes, set())



from nion.utils import Model


class Handler:

    tab_index_model = Model.PropertyModel(1)

    def switch3(self, widget):
        print("CLICKED 3")
        self.tab_index_model.value = 2


def construct_ui(ui):

    def create_tab_content(tab_label, text, check_text, button_text):
        label = ui.create_label(text=text)
        check_box = ui.create_check_box(text=check_text)
        button = ui.create_push_button(text=button_text)
        return ui.create_tab(tab_label, ui.create_column(label, check_box, button, spacing=8, margin=4))

    tab0 = create_tab_content("First", "ONE", "Check 1", "Push ONE")
    tab1 = create_tab_content("Second", "TWO", "Check 2", "Push TWO")
    tab2 = create_tab_content("Third", "THREE", "Check 3", "Push THREE")

    tabs = ui.create_tabs(tab0, tab1, tab2, current_index="@binding(tab_index_model.value)")

    button = ui.create_push_button(text="3", on_clicked="switch3")

    return ui.create_column(tabs, button)



# standard libraries
import unittest

# third party libraries
# None

# local libraries
from nion.ui import CanvasItem
from nion.ui import GridCanvasItem
from nion.utils import Geometry
from nion.utils import Selection


class GridCanvasItemDelegate:
    def __init__(self, item_count=None):
        self.__item_count = item_count if item_count is not None else 4

    @property
    def item_count(self):
        return self.__item_count

    def on_drag_started(self, mouse_index, x, y, modifiers):
        pass


class TestGridCanvasItemClass(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_shift_click_extends_selection(self):
        selection = Selection.IndexedSelection()
        delegate = GridCanvasItemDelegate()
        canvas_item = GridCanvasItem.GridCanvasItem(delegate, selection)
        canvas_item.update_layout((0, 0), (320, 100))
        self.assertEqual(selection.indexes, set())
        canvas_item.simulate_click(Geometry.IntPoint(y=120, x=50))
        self.assertEqual(selection.indexes, {1})
        modifiers = CanvasItem.KeyboardModifiers(shift=True)
        canvas_item.simulate_click(Geometry.IntPoint(y=200, x=50), modifiers)
        self.assertEqual(selection.indexes, {1, 2})

    def test_start_drag_does_not_change_selection(self):
        selection = Selection.IndexedSelection()
        delegate = GridCanvasItemDelegate()
        canvas_item = GridCanvasItem.GridCanvasItem(delegate, selection)
        canvas_item.update_layout((0, 0), (320, 100))
        self.assertEqual(selection.indexes, set())
        canvas_item.simulate_drag(Geometry.IntPoint(y=120, x=50), Geometry.IntPoint(y=120, x=500))
        self.assertEqual(selection.indexes, set())

    def test_layout_size_maintains_height_with_no_items_when_not_wrapped(self):
        selection = Selection.IndexedSelection()
        delegate = GridCanvasItemDelegate(0)
        canvas_item = GridCanvasItem.GridCanvasItem(delegate, selection, wrap=False)
        canvas_item.update_layout((0, 0), (40, 500))
        self.assertEqual(canvas_item.canvas_bounds.height, 40)



# standard libraries
import collections
import copy
import enum
import pathlib
import typing

# third party libraries
import numpy

# local libraries
from . import CanvasItem
from . import DrawingContext
from . import UserInterface as UserInterfaceModule
from nion.utils import Geometry


focused_widget = None  # simulate focus handling at the widget level


class MimeData(UserInterfaceModule.MimeData):
    def __init__(self, mime_data=None):
        self.mime_data = dict() if mime_data is None else mime_data

    @property
    def formats(self):
        return list(self.mime_data.keys())

    @property
    def file_paths(self):
        urls = self.urls
        file_paths = []
        for url in urls:
            file_paths.append(url)
        return file_paths

    def data_as_string(self, format):
        return str(self.mime_data.get(format))

    def set_data_as_string(self, format, text):
        self.mime_data[format] = text


class MenuItem(UserInterfaceModule.MenuAction):
    def __init__(self, title, action_id, callback, key_sequence, role, menu, is_separator, checked):
        super().__init__(action_id)
        self.__title = title
        self.callback = callback
        self.key_sequence = key_sequence,
        self.role = role
        self.menu = menu
        self.is_separator = is_separator
        self.__checked = checked
        self.__enabled = True
    def close(self):
        self.callback = None
        super().close()
    @property
    def title(self):
        return self.__title
    @title.setter
    def title(self, value):
        self.__title = value
    @property
    def checked(self):
        return self.__checked
    @checked.setter
    def checked(self, value):
        self.__checked = value
    @property
    def enabled(self):
        return self.__enabled
    @enabled.setter
    def enabled(self, value):
        self.__enabled = value


class Menu(UserInterfaceModule.Menu):
    def __init__(self, document_window, title=None, menu_id=None):
        super().__init__(document_window, title, menu_id)
        self.on_popup = None
    def add_menu_item(self, title: str, callback: typing.Callable[[], None], key_sequence: str = None, role: str = None, action_id: str = None) -> UserInterfaceModule.MenuAction:
        menu_item = MenuItem(title, action_id, callback, key_sequence, role, None, False, False)
        self._item_added(action=menu_item)
        return menu_item
    def add_sub_menu(self, title, menu, menu_id = None):
        menu_item = MenuItem(title, menu_id, None, None, None, menu, False, False)
        self._item_added(sub_menu=menu_item.menu)
        return menu_item
    def add_separator(self):
        self._item_added(is_separator=True)
    def popup(self, gx, gy):
        if self.on_popup:
            self.on_popup(self, gx, gy)


class ItemModelController:
    DRAG = 0
    DROP = 1
    class Item:
        def __init__(self, data=None):
            self.children = []
            self.parent = None
            self.id = None
            self.data = data if data else {}
        def insert_child(self, before_index, item):
            item.parent = self
            self.children.insert(before_index, item)
        def remove_child(self, item):
            item.parent = None
            self.children.remove(item)
        def child(self, index):
            return self.children[index]
        @property
        def row(self):
            if self.parent:
                return self.parent.children.index(self)
            return -1
    def __init__(self):
        self.__next_id = 0
        self.root = self.create_item()
    def close(self):
        pass
    def create_item(self, data=None):
        item = ItemModelController.Item(data)
        item.id = self.__next_id
        self.__next_id = self.__next_id + 1
        return item
    def item_from_id(self, item_id, parent=None):
        item = []  # nonlocal in Python 3.1+
        def fn(parent, index, child):
            if child.id == item_id:
                item.append(child)
                return True
        self.traverse(fn)
        return item[0] if item else None
    def __item_id(self, index, parent_id):
        parent = self.item_from_id(parent_id)
        assert parent is not None
        if index >= 0 and index < len(parent.children):
            return parent.children[index].id
        return 0  # invalid id
    def item_value_for_item_id(self, role, index, item_id):
        child = self.item_from_id(item_id)
        if role == "index":
            return index
        if role in child.data:
            return child.data[role]
        return None
    def item_value(self, role, index, parent_id):
        return self.item_value_for_item_id(role, index, self.__item_id(index, parent_id))
    def begin_insert(self, first_row, last_row, parent_row, parent_id):
        pass
    def end_insert(self):
        pass
    def begin_remove(self, first_row, last_row, parent_row, parent_id):
        pass
    def end_remove(self):
        pass
    def data_changed(self, row, parent_row, parent_id):
        pass
    def traverse_depth_first(self, fn, parent):
        real_parent = parent if parent else self.root
        for index, child in enumerate(real_parent.children):
            if self.traverse_depth_first(fn, child):
                return True
            if fn(parent, index, child):
                return True
        return False
    def traverse(self, fn):
        if not fn(None, 0, self.root):
            self.traverse_depth_first(fn, self.root)


class ListModelController:
    DRAG = 0
    DROP = 1
    def __init__(self):
        self.model = []
    def close(self):
        pass
    def begin_insert(self, first_row, last_row):
        pass
    def end_insert(self):
        pass
    def begin_remove(self, first_row, last_row):
        pass
    def end_remove(self):
        pass
    def data_changed(self):
        pass


class Key(UserInterfaceModule.Key):
    def __init__(self, text, key, modifiers):
        self.__text = text
        self.__key = key
        self.__modifiers = modifiers if modifiers else CanvasItem.KeyboardModifiers()

    @property
    def text(self) -> str:
        return self.__text

    @property
    def key(self) -> str:
        return self.__key

    @property
    def modifiers(self) -> UserInterfaceModule.KeyboardModifiers:
        return self.__modifiers

    @property
    def is_delete(self):
        return self.key == "delete"

    @property
    def is_enter_or_return(self):
        return self.key == "enter" or self.key == "return"

    @property
    def is_tab(self):
        return self.key == "tab"

    @property
    def is_arrow(self):
        return self.is_left_arrow or self.is_up_arrow or self.is_right_arrow or self.is_down_arrow

    @property
    def is_left_arrow(self):
        return self.key == "left"

    @property
    def is_up_arrow(self):
        return self.key == "up"

    @property
    def is_right_arrow(self):
        return self.key == "right"

    @property
    def is_down_arrow(self):
        return self.key == "down"

    @property
    def is_delete_to_end_of_line(self):
        return self.key == "delete_to_end_of_line"

    @property
    def is_end(self):
        return self.key == "end"

    @property
    def is_escape(self):
        return self.key == "escape"

    @property
    def is_home(self):
        return self.key == "home"

    @property
    def is_insert(self):
        return self.key == "insert"

    @property
    def is_move_to_end_of_line(self):
        return self.key == "end_of_line"

    @property
    def is_move_to_start_of_line(self):
        return self.key == "start_of_line"

    @property
    def is_page_down(self):
        return self.key == "page_down"

    @property
    def is_page_up(self):
        return self.key == "page_up"


class ButtonGroup:

    def __init__(self):
        self.on_button_clicked = None

    def close(self):
        self.on_button_clicked = None

    def add_button(self, radio_button, button_id):
        pass

    def remove_button(self, radio_button):
        pass

    def clicked(self, button_id):
        if self.on_button_clicked:
            self.on_button_clicked(button_id)


class Widget:

    def __init__(self, widget_type: str):
        self.widget_type = widget_type
        self.children: typing.List[Widget] = list()
        self.size: typing.Optional[int] = None
        self.on_size_changed: typing.Optional[typing.Callable[[Geometry.IntSize], None]] = None

    def close(self) -> None:
        self.children = typing.cast(typing.List[Widget], None)

    def size_changed(self, size) -> None:
        if size != self.size:
            self.size = size
            if callable(self.on_size_changed):
                self.on_size_changed(size)
        for child in self.children:
            child.size_changed(size)


class WidgetBehavior:

    def __init__(self, widget_type: str, properties: typing.Mapping):
        self.widget = Widget(widget_type)
        self.widget.on_size_changed = self._size_changed
        self.does_retain_focus = False
        self.on_ui_activity = None
        self.on_context_menu_event = None
        self.on_focus_changed = None
        self.on_size_changed = None
        self._no_focus = "no_focus"
        self.__focused = False
        self.visible = True
        self.enabled = True
        self.size = None
        self.tool_tip = None
        self.children: typing.List[Widget] = list()
        self.content = None
        self.canvas_item = None

    def close(self) -> None:
        if callable(getattr(self.widget, "close", None)):
            self.widget.close()
        self.on_ui_activity = None
        self.on_context_menu_event = None
        self.on_focus_changed = None
        self.on_size_changed = None
        self.children = list()
        self.content = None
        self.canvas_item = None

    def _set_root_container(self, root_container):
        pass

    def _register_ui_activity(self):
        pass

    @property
    def focused(self) -> bool:
        return self.__focused

    @focused.setter
    def focused(self, focused: bool) -> None:
        global focused_widget
        if self.__focused != focused:
            if focused and focused_widget != self:
                if focused_widget:
                    focused_widget.focused = False
                focused_widget = self
            self.__focused = focused
            if self.on_focus_changed:
                self.on_focus_changed(focused)

    def set_property(self, key: str, value) -> None:
        pass

    def _size_changed(self, size: Geometry.IntSize) -> None:
        pass

    def map_to_global(self, p):
        return p


class NullBehavior:

    def __init__(self):
        self.focused = False
        self.enabled = True
        self.visible = True

    def close(self):
        pass

    def _set_root_container(self, root_container):
        pass


class BoxStretch(UserInterfaceModule.Widget):

    def __init__(self):
        super().__init__(NullBehavior())


class BoxSpacing(UserInterfaceModule.Widget):

    def __init__(self, spacing: int):
        super().__init__(NullBehavior())
        self.spacing = spacing


def extract_widget(widget: UserInterfaceModule.Widget) -> typing.Optional[Widget]:
    content_widget = getattr(widget, "content_widget", None)
    if content_widget:
        return extract_widget(content_widget)
    elif hasattr(widget, "_behavior"):
        return widget._behavior.widget
    return None


class WidgetItemType(enum.Enum):
    WIDGET = 0
    STRETCH = 1
    SPACING = 2


class WidgetItem:

    def __init__(self, type: WidgetItemType, *, widget=None, fill: bool = False, alignment: str = None, spacing: int = 0):
        self.type = type
        self.widget = widget
        self.fill = fill
        self.alignment = alignment
        self.spacing = spacing


class BoxWidgetBehavior(WidgetBehavior):

    def __init__(self, widget_type: str, properties: typing.Mapping):
        super().__init__(widget_type, properties)
        self.__widgets: typing.List[WidgetItem] = list()

    def insert(self, child: UserInterfaceModule.Widget, index_or_widget: typing.Optional[typing.Union[UserInterfaceModule.Widget, int]], fill: bool = False, alignment: typing.Optional[str] = None) -> None:
        # behavior must handle index of None, meaning insert at end
        child_widget = extract_widget(child)
        assert child_widget is not None
        index = index_or_widget if isinstance(index_or_widget, int) else len(self.__widgets)
        self.__widgets.insert(index, WidgetItem(WidgetItemType.WIDGET, widget=child_widget, fill=fill, alignment=alignment))
        self.widget.children.insert(index, child_widget)
        child_widget.size_changed(self.widget.size)

    def add_stretch(self) -> UserInterfaceModule.Widget:
        self.__widgets.append(WidgetItem(WidgetItemType.STRETCH))
        self.widget.children.append(Widget("stretch"))
        return BoxStretch()

    def add_spacing(self, spacing: int) -> UserInterfaceModule.Widget:
        self.__widgets.append(WidgetItem(WidgetItemType.SPACING, spacing=spacing))
        self.widget.children.append(Widget("spacing"))
        return BoxSpacing(spacing)

    def remove_all(self) -> None:
        self.__widgets.clear()
        self.widget.children.clear()


class SplitterWidgetBehavior(WidgetBehavior):

    def __init__(self, widget_type: str, properties: typing.Mapping):
        super().__init__(widget_type, properties)

    def add(self, child: UserInterfaceModule.Widget) -> None:
        # behavior must handle index of None, meaning insert at end
        child_widget = extract_widget(child)
        assert child_widget is not None
        index = len(self.widget.children)
        self.widget.children.insert(index, child_widget)
        child_widget.size_changed(self.widget.size)

    def restore_state(self, tag: str) -> None:
        pass

    def save_state(self, tag: str) -> None:
        pass

    def set_sizes(self, sizes: typing.Sequence[int]) -> None:
        for child, size in zip(self.widget.children, sizes):
            child.size = size


class TabWidgetBehavior(WidgetBehavior):

    def __init__(self, widget_type: str, properties: typing.Mapping):
        super().__init__(widget_type, properties)


class StackWidgetBehavior(WidgetBehavior):

    def __init__(self, widget_type: str, properties: typing.Mapping):
        super().__init__(widget_type, properties)
        self.current_index = -1

    def insert(self, child: UserInterfaceModule.Widget, index: int) -> None:
        # behavior must handle index of None, meaning insert at end
        child_widget = extract_widget(child)
        assert child_widget is not None
        index = index if index is not None else len(self.widget.children)
        self.widget.children.insert(index, child_widget)
        child_widget.size_changed(self.widget.size)

    def add(self, child: UserInterfaceModule.Widget) -> None:
        child_widget = extract_widget(child)
        assert child_widget
        self.widget.children.append(child_widget)

    def remove(self, child: UserInterfaceModule.Widget) -> None:
        child_widget = extract_widget(child)
        assert child_widget
        self.widget.children.remove(child_widget)


class GroupWidgetBehavior(WidgetBehavior):

    def __init__(self, widget_type: str, properties: typing.Mapping):
        super().__init__(widget_type, properties)


class ScrollAreaWidgetBehavior(WidgetBehavior):

    def __init__(self, widget_type: str, properties: typing.Mapping):
        super().__init__(widget_type, properties)
        self.on_size_changed = None
        self.on_viewport_changed = None

    def close(self):
        self.on_size_changed = None
        self.on_viewport_changed = None
        super().close()

    def set_content(self, content: UserInterfaceModule.Widget) -> None:
        assert not self.widget.children
        child_widget = extract_widget(content)
        assert child_widget
        self.widget.children.append(child_widget)
        child_widget.size_changed(self.widget.size)

    # called from widget
    def _size_changed(self, size: Geometry.IntSize) -> None:
        self._register_ui_activity()
        if callable(self.on_size_changed):
            self.on_size_changed(size.width, size.height)

    def scroll_to(self, x, y):
        pass

    def set_scrollbar_policies(self, horizontal_policy, vertical_policy):
        pass


class ComboBoxWidgetBehavior(WidgetBehavior):

    def __init__(self, widget_type: str, properties: typing.Mapping):
        super().__init__(widget_type, properties)
        self.on_current_text_changed = None
        self.current_index = 0
        self.item_strings: typing.List[str] = list()

    def close(self):
        self.on_current_text_changed = None
        super().close()

    @property
    def current_text(self) -> str:
        return self.item_strings[self.current_index] if 0 <= self.current_index < len(self.item_strings) else str()

    @current_text.setter
    def current_text(self, value: str) -> None:
        if 0 <= self.current_index < len(self.item_strings):
            self.item_strings[self.current_index] = value

    def set_item_strings(self, item_strings: typing.Sequence[str]) -> None:
        self.item_strings = list(item_strings)


class PushButtonWidgetBehavior(WidgetBehavior):

    def __init__(self, widget_type: str, properties: typing.Mapping):
        super().__init__(widget_type, properties)


class RadioButtonWidgetBehavior(WidgetBehavior):

    def __init__(self, widget_type: str, properties: typing.Mapping):
        super().__init__(widget_type, properties)


class CheckBoxWidgetBehavior(WidgetBehavior):

    def __init__(self, widget_type: str, properties: typing.Mapping):
        super().__init__(widget_type, properties)


class LabelWidgetBehavior(WidgetBehavior):

    def __init__(self, widget_type: str, properties: typing.Mapping):
        super().__init__(widget_type, properties)

    def set_text_color(self, color: str) -> None:
        pass

    def set_text_font(self, font_str: str) -> None:
        pass


class SliderWidgetBehavior(WidgetBehavior):

    def __init__(self, widget_type: str, properties: typing.Mapping):
        super().__init__(widget_type, properties)
        self.value = 0


class CanvasWidgetBehavior(WidgetBehavior):

    def __init__(self, widget_type: str, properties: typing.Mapping):
        super().__init__(widget_type, properties)
        self.on_mouse_entered = None
        self.on_mouse_exited = None
        self.on_mouse_clicked = None
        self.on_mouse_double_clicked = None
        self.on_mouse_pressed = None
        self.on_mouse_released = None
        self.on_mouse_position_changed = None
        self.on_grabbed_mouse_position_changed = None
        self.on_wheel_changed = None
        self.on_key_pressed = None
        self.on_key_released = None
        self.on_size_changed = None
        self.on_drag_enter = None
        self.on_drag_leave = None
        self.on_drag_move = None
        self.on_drop = None
        self.on_tool_tip = None
        self.on_pan_gesture = None
        self.__focusable = False

    def close(self):
        self.on_mouse_entered = None
        self.on_mouse_exited = None
        self.on_mouse_clicked = None
        self.on_mouse_double_clicked = None
        self.on_mouse_pressed = None
        self.on_mouse_released = None
        self.on_mouse_position_changed = None
        self.on_grabbed_mouse_position_changed = None
        self.on_wheel_changed = None
        self.on_key_pressed = None
        self.on_key_released = None
        self.on_size_changed = None
        self.on_drag_enter = None
        self.on_drag_leave = None
        self.on_drag_move = None
        self.on_drop = None
        self.on_tool_tip = None
        self.on_pan_gesture = None
        super().close()

    def _set_canvas_item(self, canvas_item):
        pass

    def periodic(self):
        pass

    @property
    def focusable(self):
        return self.__focusable

    @focusable.setter
    def focusable(self, focusable):
        self.__focusable = focusable

    def draw(self, drawing_context):
        pass

    def draw_section(self, section_id: int, drawing_context: DrawingContext.DrawingContext, canvas_rect: Geometry.IntRect) -> None:
        pass

    def remove_section(self, section_id: int) -> None:
        pass

    def set_cursor_shape(self, cursor_shape):
        cursor_shape = cursor_shape or "arrow"

    def grab_gesture(self, gesture_type):
        pass

    def release_gesture(self, gesture_type):
        pass

    def grab_mouse(self, gx, gy):
        pass

    def release_mouse(self):
        pass

    def show_tool_tip_text(self, text: str, gx: int, gy: int) -> None:
        pass


class LineEditWidgetBehavior(WidgetBehavior):

    def __init__(self, widget_type: str, properties: typing.Mapping):
        super().__init__(widget_type, properties)
        self.on_editing_finished = None
        self.on_escape_pressed = None
        self.on_return_pressed = None
        self.on_key_pressed = None
        self.on_text_edited = None
        self.__clear_button_enabled = False
        self._no_focus = "click_focus"
        self.text = str()
        self.placeholder_text = str()
        self.editable = True
        self.clear_button_enabled = False

    def close(self):
        self.on_editing_finished = None
        self.on_escape_pressed = None
        self.on_return_pressed = None
        self.on_text_edited = None
        super().close()

    @property
    def selected_text(self) -> str:
        return self.text

    def select_all(self) -> None:
        pass

    def editing_finished(self, text: str) -> None:
        self.text = text
        if self.on_editing_finished:
            self.on_editing_finished(text)


class TextEditWidgetBehavior(WidgetBehavior):

    def __init__(self, widget_type: str, properties: typing.Mapping):
        super().__init__(widget_type, properties)
        self.__word_wrap_mode = "optimal"
        self.on_cursor_position_changed = None
        self.on_selection_changed = None
        self.on_text_changed = None
        self.on_escape_pressed = None
        self.on_return_pressed = None
        self.on_key_pressed = None
        self.on_insert_mime_data = None
        self._no_focus = "click_focus"
        self.text = str()
        self.selected_text = str()
        self.placeholder = str()
        self.editable = True
        self.word_wrap_mode = None

    def close(self):
        self.on_cursor_position_changed = None
        self.on_selection_changed = None
        self.on_text_changed = None
        self.on_escape_pressed = None
        self.on_return_pressed = None
        self.on_key_pressed = None
        self.on_insert_mime_data = None
        super().close()

    @property
    def cursor_position(self) -> UserInterfaceModule.CursorPosition:
        return UserInterfaceModule.CursorPosition(0, 0, 0)

    @property
    def selection(self) -> UserInterfaceModule.Selection:
        return UserInterfaceModule.Selection(0, 0)

    def append_text(self, value: str) -> None:
        self.text += value

    def insert_text(self, value: str) -> None:
        self.text += value

    def clear_selection(self) -> None:
        pass

    def remove_selected_text(self) -> None:
        pass

    def select_all(self) -> None:
        pass

    def move_cursor_position(self, operation, mode=None, n: int = 1) -> None:
        pass

    def set_line_height_proportional(self, proportional_line_height: float) -> None:
        pass

    def set_text_background_color(self, color: str) -> None:
        pass

    def set_text_color(self, color: str) -> None:
        pass

    def set_text_font(self, font_str: str) -> None:
        pass


class TreeWidgetBehavior(WidgetBehavior):

    def __init__(self, widget_type: str, properties: typing.Mapping):
        super().__init__(widget_type, properties)

    def set_current_row(self, index, parent_row, parent_id):
        pass

    def clear_current_row(self):
        pass

    def size_to_content(self):
        pass


class DocumentWindowX(UserInterfaceModule.Window):

    def __init__(self, size: typing.Optional[Geometry.IntSize] = None):
        super().__init__(None, "title")
        self.__size = size if size is not None else Geometry.IntSize(height=720, width=960)
        self.__title = None

    def request_close(self):
        if self.on_about_to_close:
            self.on_about_to_close(str(), str())

    def _attach_root_widget(self, root_widget):
        extract_widget(root_widget).size_changed(self.__size)

    def _set_title(self, value: str) -> None:
        self.__title = value

    def _set_window_file_path(self, value: typing.Optional[pathlib.Path]) -> None:
        self.__window_file_path = value

    def set_palette_color(self, role: str, r: int, g: int, b: int, a: int) -> None:
        pass

    def set_window_style(self, styles: typing.Sequence[str]) -> None:
        pass

    def set_attributes(self, attributes: typing.Sequence[str]) -> None:
        pass

    @property
    def position(self) -> Geometry.IntPoint:
        return Geometry.IntPoint(x=0, y=0)

    @property
    def size(self) -> Geometry.IntSize:
        return Geometry.IntSize(w=640, h=480)

    def create_dock_widget(self, widget: UserInterfaceModule.Widget, panel_id: str, title: str, positions: typing.Sequence[str], position: str) -> UserInterfaceModule.DockWidget:
        dock_widget = DockWidget(self, widget, panel_id, title, positions, position)
        dock_widget.size_changed(Geometry.IntSize(height=320, width=480))
        return dock_widget

    def tabify_dock_widgets(self, dock_widget1, dock_widget2):
        pass

    def insert_menu(self, title: str, before_menu, menu_id: str = None) -> UserInterfaceModule.Menu:
        menu = Menu(self)
        self._menu_inserted(menu, before_menu)
        return menu

    def add_menu(self, title: str, menu_id: str = None) -> Menu:
        menu = Menu(self, menu_id)
        self._menu_added(menu)
        return menu

    def show(self, size=None, position=None):
        pass

    def restore(self, geometry, state):
        pass

    def _get_focus_widget(self):
        global focused_widget
        return focused_widget


class DockWidget(UserInterfaceModule.DockWidget):

    def __init__(self, window, widget: UserInterfaceModule.Widget, panel_id: str, title: str, positions: typing.Sequence[str], position: str):
        super().__init__(window, widget, panel_id, title, positions, position)
        self.visible = False
        self.__focus_policy = "no_focus"
        self.does_retain_focus = False

    @property
    def toggle_action(self):
        action = UserInterfaceModule.MenuAction("toggle_dock_widget_" + self.panel_id)
        action.on_ui_activity = self._register_ui_activity
        return action

    def show(self):
        self.visible = True
        self._register_ui_activity()

    def hide(self):
        self.visible = False
        self._register_ui_activity()

    def size_changed(self, size: Geometry.IntSize) -> None:
        self._handle_size_changed(size)


class UserInterface(UserInterfaceModule.UserInterface):

    def __init__(self):
        CanvasItem._threaded_rendering_enabled = False
        self.clipboard = MimeData()
        self.popup = None
        self.popup_pos = None

    def close(self):
        pass

    def request_quit(self) -> None:
        pass

    def set_application_info(self, name: str, organization: str, domain: str) -> None:
        pass

    # data objects

    def create_mime_data(self) -> MimeData:
        return MimeData()

    def create_item_model_controller(self, keys):
        return ItemModelController()

    def create_button_group(self):
        return ButtonGroup()

    # window elements

    def create_document_window(self, title=None, parent_window=None):
        return DocumentWindowX()

    def destroy_document_window(self, document_window):
        document_window.close()

    def create_row_widget(self, alignment=None, properties=None):
        return UserInterfaceModule.BoxWidget(BoxWidgetBehavior("row", properties), alignment)

    def create_column_widget(self, alignment=None, properties=None):
        return UserInterfaceModule.BoxWidget(BoxWidgetBehavior("column", properties), alignment)

    def create_splitter_widget(self, orientation="vertical", properties=None):
        return UserInterfaceModule.SplitterWidget(SplitterWidgetBehavior("splitter", properties), orientation)

    def create_tab_widget(self, properties=None):
        return UserInterfaceModule.TabWidget(TabWidgetBehavior("tab", properties))

    def create_stack_widget(self, properties=None):
        return UserInterfaceModule.StackWidget(StackWidgetBehavior("stack", properties))

    def create_group_widget(self, properties=None):
        return UserInterfaceModule.GroupWidget(GroupWidgetBehavior("group", properties))

    def create_scroll_area_widget(self, properties=None):
        return UserInterfaceModule.ScrollAreaWidget(ScrollAreaWidgetBehavior("scrollarea", properties))

    def create_combo_box_widget(self, items=None, item_getter=None, properties=None):
        return UserInterfaceModule.ComboBoxWidget(ComboBoxWidgetBehavior("combobox", properties), items, item_getter)

    def create_push_button_widget(self, text=None, properties=None):
        return UserInterfaceModule.PushButtonWidget(PushButtonWidgetBehavior("pushbutton", properties), text)

    def create_radio_button_widget(self, text=None, properties=None):
        return UserInterfaceModule.RadioButtonWidget(RadioButtonWidgetBehavior("radiobutton", properties), text)

    def create_check_box_widget(self, text=None, properties=None):
        return UserInterfaceModule.CheckBoxWidget(CheckBoxWidgetBehavior("checkbox", properties), text)

    def create_label_widget(self, text=None, properties=None):
        return UserInterfaceModule.LabelWidget(LabelWidgetBehavior("label", properties), text)

    def create_slider_widget(self, properties=None):
        return UserInterfaceModule.SliderWidget(SliderWidgetBehavior("slider", properties))

    def create_progress_bar_widget(self, properties=None):
        return UserInterfaceModule.ProgressBarWidget(CanvasWidgetBehavior("canvas", properties))

    def create_line_edit_widget(self, properties=None):
        return UserInterfaceModule.LineEditWidget(LineEditWidgetBehavior("lineedit", properties))

    def create_text_edit_widget(self, properties=None):
        return UserInterfaceModule.TextEditWidget(TextEditWidgetBehavior("textedit", properties))

    def create_canvas_widget(self, properties=None, *, layout_render: str = None):
        return UserInterfaceModule.CanvasWidget(CanvasWidgetBehavior("canvas", properties), layout_render=layout_render)

    def create_tree_widget(self, properties=None):
        return UserInterfaceModule.TreeWidget(TreeWidgetBehavior("pytree", properties))

    # file i/o

    def load_rgba_data_from_file(self, filename):
        return numpy.zeros((20,20), numpy.uint32)

    def save_rgba_data_to_file(self, data, filename, format):
        pass

    def get_existing_directory_dialog(self, title, directory):
        return directory, directory

    def get_file_paths_dialog(self, title: str, directory: str, filter: str, selected_filter: str=None) -> typing.Tuple[typing.List[str], str, str]:
        raise NotImplementedError()

    def get_file_path_dialog(self, title, directory, filter, selected_filter=None):
        raise NotImplementedError()

    def get_save_file_path(self, title, directory, filter, selected_filter=None):
        raise NotImplementedError()

    # persistence (associated with application)

    def get_data_location(self):
        return str()

    def get_document_location(self):
        return str()

    def get_temporary_location(self):
        return str()

    def get_configuration_location(self):
        return str()

    def get_persistent_string(self, key, default_value=None):
        return default_value

    def set_persistent_string(self, key, value):
        pass

    def get_persistent_object(self, key, default_value=None):
        return default_value

    def set_persistent_object(self, key, value):
        pass

    def remove_persistent_key(self, key):
        pass

    # clipboard

    def clipboard_clear(self):
        self.clipboard = MimeData()

    def clipboard_mime_data(self) -> MimeData:
        return self.clipboard

    def clipboard_set_mime_data(self, mime_data: UserInterfaceModule.MimeData) -> None:
        self.clipboard = mime_data

    def clipboard_set_text(self, text):
        self.clipboard = MimeData()
        self.clipboard.set_data_as_string('text', text)

    def clipboard_text(self):
        self.clipboard.data_as_string('text')

    # misc

    def create_rgba_image(self, drawing_context, width, height):
        return numpy.zeros((height, width), dtype=numpy.uint32)

    def get_font_metrics(self, font_str: str, text: str) -> UserInterfaceModule.FontMetrics:
        return UserInterfaceModule.FontMetrics(width=(len(text) * 12), height=12, ascent=10, descent=2, leading=0)

    def truncate_string_to_width(self, font_str: str, text: str, pixel_width: int, mode: UserInterfaceModule.TruncateModeType) -> str:
        return text

    def get_qt_version(self) -> str:
        return "TEST"

    def get_tolerance(self, tolerance_type: UserInterfaceModule.ToleranceType) -> float:
        return 5

    def create_context_menu(self, document_window) -> UserInterfaceModule.Menu:
        menu = Menu(document_window)
        def handle_popup(menu, gx, gy):
            self.popup = menu
            self.popup_pos = gx, gy
        menu.on_popup = handle_popup
        return menu

    def create_sub_menu(self, document_window, title: str = None, menu_id: str = None) -> UserInterfaceModule.Menu:
        return Menu(document_window, title)

    def get_color_dialog(self, title: str, color: typing.Optional[str], show_alpha: bool) -> typing.Optional[str]:
        return color

    def get_keyboard_modifiers(self, query: bool = False) -> UserInterfaceModule.KeyboardModifiers:
        return typing.cast(UserInterfaceModule.KeyboardModifiers, CanvasItem.KeyboardModifiers())

    # testing

    def create_key_by_id(self, key_id: str, modifiers: CanvasItem.KeyboardModifiers = None) -> UserInterfaceModule.Key:
        return Key(None, key_id, modifiers)

    def create_modifiers_by_id_list(self, modifiers_id_list: typing.Sequence[str]) -> CanvasItem.KeyboardModifiers:
        shift = False
        control = False
        alt = False
        meta = False
        keypad = False
        for modifiers_id in modifiers_id_list:
            if modifiers_id == "shift":
                shift = True
            elif modifiers_id == "control":
                control = True
            elif modifiers_id == "alt":
                alt = True
            elif modifiers_id == "meta":
                meta = True
            elif modifiers_id == "keypad":
                keypad = True
        return CanvasItem.KeyboardModifiers(shift, control, alt, meta, keypad)



import typing
from nion.ui import Declarative
from nion.utils import ListModel
from nion.utils import Model


class Shape:
    def __init__(self):
        self.type = "shape"
        self.label = "Shape"

class Rectangle(Shape):
    def __init__(self):
        super().__init__()
        self.type = "rectangle"
        self.label = "Rectangle"
        self.width = 100
        self.height = 90


class Circle(Shape):
    def __init__(self):
        super().__init__()
        self.type = "circle"
        self.label = "Circle"
        self.radius = 45


class Interval(Shape):
    def __init__(self):
        super().__init__()
        self.type = "interval"
        self.label = "Interval"
        self.left = 55
        self.right = 70


class ShapeHandler:

    def __init__(self, shape: Shape):
        self.shape = shape


class Handler:

    def __init__(self):
        # define our list of shapes, one from each class
        self.shapes_model = ListModel.ListModel(items=[Rectangle(), Circle(), Interval()])

        # define the shape index model
        self.shape_index_model = Model.PropertyModel(0)

        # define the shape page model
        self.shape_page = Model.PropertyModel()

        def shape_index_changed(p: str = None) -> None:
            self.shape_page.value = self.shapes_model.items[self.shape_index_model.value].type

        shape_index_changed()

        # track changes to shape index and update shape page.
        self.__shape_index_changed = self.shape_index_model.property_changed_event.listen(shape_index_changed)

    def close(self):
        self.__shape_index_changed.close()
        self.__shape_index_changed = None

    def create_handler(self, component_id: str, container=None, item=None, **kwargs):
        if component_id in ("shape", "rectangle", "circle", "interval"):
            item = self.shapes_model.items[self.shape_index_model.value]
            return ShapeHandler(item)
        return None

    def get_resource(self, resource_id: str, container=None, item=None) -> typing.Optional[Declarative.UIDescription]:
        u = Declarative.DeclarativeUI()
        if resource_id == "rectangle":
            width_row = u.create_row(u.create_label(text="Width:"), u.create_label(text="@binding(shape.width)"), spacing=8)
            height_row = u.create_row(u.create_label(text="Height:"), u.create_label(text="@binding(shape.height)"), spacing=8)
            return u.define_component(u.create_group(u.create_column(width_row, height_row, spacing=8), title="@binding(shape.label)"))
        elif resource_id == "circle":
            radius_row = u.create_row(u.create_label(text="Radius:"), u.create_label(text="@binding(shape.radius)"), spacing=8)
            return u.define_component(u.create_group(u.create_column(radius_row, spacing=8), title="@binding(shape.label)"))
        elif resource_id == "interval":
            left_row = u.create_row(u.create_label(text="Left:"), u.create_label(text="@binding(shape.left)"), spacing=8)
            right_row = u.create_row(u.create_label(text="Right:"), u.create_label(text="@binding(shape.right)"), spacing=8)
            return u.define_component(u.create_group(u.create_column(left_row, right_row, spacing=8), title="@binding(shape.label)"))
        return None


def construct_ui(ui: Declarative.DeclarativeUI):
    shape_choice = ui.create_combo_box(items=["Rectangle", "Circle", "Interval"], current_index="@binding(shape_index_model.value)")
    shape_component = ui.create_component_instance("@binding(shape_page.value)")
    return ui.create_column(shape_choice, shape_component, spacing=8, margin=12)



"""
A basic class to serve as the basis of a typical one window application.
"""
from __future__ import annotations

# standard libraries
import asyncio
import copy
import gettext
import logging
import os
import sys
import typing
import weakref

# local libraries
from nion.ui import Declarative
from nion.ui import UserInterface
from nion.ui import Window
from nion.utils import Process

if typing.TYPE_CHECKING:
    from nion.utils import Event

_ = gettext.gettext


class LoggingHandler(logging.StreamHandler):

    def __init__(self):
        super().__init__()
        self.__records = list()

    def emit(self, record):
        super().emit(record)
        if self.__records is not None:
            self.__records.append(record)

    def take_records(self):
        records = self.__records
        self.__records = None
        return records


logging_handler = LoggingHandler()


class BaseApplication:
    """A basic application class.

    Subclass this class and implement the start method. The start method should create a document window that will be
    the focus of the UI.

    Pass the desired user interface to the init method. Then call initialize and start.
    """

    def __init__(self, ui: UserInterface.UserInterface, *, on_start=None):
        self.ui = ui

        # handle last window closing in Python; but use this variable so tool can continue to
        # close on last window for backwards compatibility. setting this flag will tell the
        # tool to not close on last window closing. it is handled when items are removed from
        # __windows instead.
        self._should_close_on_last_window = False
        self.__prevent_close_count = 0

        self.__on_start = on_start
        logger = logging.getLogger()
        logger.setLevel(logging.INFO)
        logger.addHandler(logging_handler)

        class ContextFilter(logging.Filter):
            def filter(self, record):
                if record.exc_info and record.exc_info[0] == ConnectionResetError:
                    return False
                return True

        # work around bug 39010 here
        logger = logging.getLogger('asyncio')
        logger.addFilter(ContextFilter())

        self.__windows : typing.List[Window.Window] = list()
        self.__window_close_event_listeners: typing.Dict[Window.Window, Event.EventListener] = dict()
        self.__event_loop: asyncio.AbstractEventLoop = typing.cast(asyncio.AbstractEventLoop, None)
        self.__dialogs : typing.List[weakref.ReferenceType] = list()

    def initialize(self):
        """Initialize. Separate from __init__ so that overridden methods can be called."""
        # configure the event loop, which can be used for non-window clients.
        logger = logging.getLogger()
        old_level = logger.level
        logger.setLevel(logging.INFO)
        self.__event_loop = asyncio.new_event_loop()  # outputs a debugger message!
        logger.setLevel(old_level)

    def deinitialize(self):
        self._close_dialogs()
        Process.close_event_loop(self.__event_loop)
        self.__event_loop = typing.cast(asyncio.AbstractEventLoop, None)
        with open(os.path.join(self.ui.get_data_location(), "PythonConfig.ini"), 'w') as f:
            f.write(sys.prefix + '\n')
        self.ui.close()

    def run(self):
        """Run the application. Called from PyQt."""
        self.ui.run(self)

    def start(self):
        """The start method should create a window that will be the focus of the UI."""
        if callable(self.__on_start):
            return self.__on_start()

    def stop(self):
        # program is really stopping, clean up.
        self.deinitialize()

    @property
    def windows(self) -> typing.List[Window.Window]:
        return copy.copy(self.__windows)

    def _window_created(self, window: Window.Window) -> None:
        self.__window_close_event_listeners[window] = window._window_close_event.listen(self._window_did_close)
        assert window not in self.__windows
        self.__windows.append(window)

    def _window_did_close(self, window: Window.Window) -> None:
        self.__window_close_event_listeners[window].close()
        del self.__window_close_event_listeners[window]
        self.__windows.remove(window)
        with self.prevent_close():
            pass  # trigger close if all windows closed.

    def prevent_close(self):

        class Context:
            def __init__(self, application: BaseApplication):
                self.__application = application

            def __enter__(self):
                self.__application._enter_prevent_close_state()
                return self

            def __exit__(self, type, value, traceback):
                self.__application._exit_prevent_close_state()

        return Context(self)

    def _enter_prevent_close_state(self) -> None:
        self.__prevent_close_count += 1

    def _exit_prevent_close_state(self) -> None:
        self.__prevent_close_count -= 1
        if not self.__prevent_close_count and not self.__windows and not self.__dialogs:
            self.ui.request_quit()

    def exit(self):
        """The exit method should request to close or close the window."""
        for window in copy.copy(self.__windows):
            # closing the window will trigger the about_to_close event to be called which
            # will then call window close which will fire its _window_close_event which will
            # remove the window from the list of window in _window_did_close.
            window.request_close()

    def periodic(self):
        """The periodic method can be overridden to implement periodic behavior."""
        if self.__event_loop:  # special for shutdown
            self.__event_loop.stop()
            self.__event_loop.run_forever()

    def _close_dialogs(self) -> None:
        for weak_dialog in self.__dialogs:
            dialog = typing.cast(Window.Window, weak_dialog())
            if dialog:
                try:
                    dialog.request_close()
                except Exception as e:
                    pass
        self.__dialogs = list()

    def is_dialog_type_open(self, dialog_class) -> bool:
        for dialog_weakref in self.__dialogs:
            if isinstance(dialog_weakref(), dialog_class):
                return True
        return False

    def register_dialog(self, dialog: Window.Window) -> None:
        def close_dialog():
            self.__dialogs.remove(weakref.ref(dialog))
        dialog.on_close = close_dialog
        self.__dialogs.append(weakref.ref(dialog))

    @property
    def event_loop(self) -> asyncio.AbstractEventLoop:
        return self.__event_loop

    def _menu_about_to_show(self, window: Window.Window, menu: UserInterface.Menu) -> bool:
        return False

    def show_ok_dialog(self, title: str, message: str, *, completion_fn: typing.Optional[typing.Callable[[], None]] = None) -> None:
        u = Declarative.DeclarativeUI()
        error_message = u.create_label(text=message)
        button_row = u.create_row(u.create_stretch(), u.create_push_button(text=_("OK"), on_clicked="close_window"))
        main_column = u.create_column(error_message, button_row, spacing=8, width=300)
        window = u.create_window(main_column, title=title, margin=12, window_style="tool")
        Declarative.WindowHandler(completion_fn=completion_fn).run(window, app=self)

    def show_ok_cancel_dialog(self, title: str, message: str, *, ok_text: str = None, cancel_text: str = None, completion_fn: typing.Optional[typing.Callable[[bool], None]] = None) -> None:
        u = Declarative.DeclarativeUI()
        error_message = u.create_label(text=message)
        button_row = u.create_row(u.create_stretch(),
                                  u.create_push_button(text=cancel_text or _("Cancel"), on_clicked="handle_reject"),
                                  u.create_push_button(text=ok_text or _("OK"), on_clicked="handle_accept"),
                                  spacing=12)
        main_column = u.create_column(error_message, button_row, spacing=8, width=380)
        window = u.create_window(main_column, title=title, margin=12, window_style="tool")

        class OkCancelHandler(Declarative.WindowHandler):
            def __init__(self):
                super().__init__(completion_fn=self.handle_close)
                self.__result = False

            def handle_close(self) -> None:
                if callable(completion_fn):
                    completion_fn(self.__result)

            def handle_accept(self, widget: typing.Optional[Declarative.UIWidget]) -> None:
                self.__result = True
                self.close_window(widget)

            def handle_reject(self, widget: typing.Optional[Declarative.UIWidget]) -> None:
                self.__result = False
                self.close_window(widget)

        OkCancelHandler().run(window, app=self)


def make_ui(bootstrap_args):
    if "proxy" in bootstrap_args:
        from nion.ui import QtUserInterface
        proxy = bootstrap_args["proxy"]
        return QtUserInterface.QtUserInterface(proxy)
    elif "pyqt" in bootstrap_args:
        from nion.ui import QtUserInterface
        from nion.ui import PyQtProxy
        return QtUserInterface.QtUserInterface(PyQtProxy.PyQtProxy())
    elif "server" in bootstrap_args:
        from nion.ui import CanvasUI
        server = bootstrap_args["server"]
        return CanvasUI.CanvasUserInterface(server.draw, server.get_font_metrics)
    else:
        return None


def run_window(args, bootstrap_args, d, handler) -> BaseApplication:
    """Make base application and run it with the declarative d and handler."""

    def start():
        Declarative.run_window(d, handler, app=app)
        return True

    app = BaseApplication(make_ui(bootstrap_args), on_start=start)
    app.initialize()

    return app
